"""Factory for serializing messages according to WolkAbout Protocol."""
#   Copyright 2020 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import json
from time import time
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

from wolk import logger_factory
from wolk.interface.message_factory import MessageFactory
from wolk.model.data_type import DataType
from wolk.model.feed_type import FeedType
from wolk.model.file_management_status import FileManagementStatus
from wolk.model.file_management_status_type import FileManagementStatusType
from wolk.model.firmware_update_status import FirmwareUpdateStatus
from wolk.model.firmware_update_status_type import FirmwareUpdateStatusType
from wolk.model.message import Message
from wolk.model.unit import Unit

OutgoingDataTypes = Union[bool, int, float, str]
Reading = Tuple[str, OutgoingDataTypes]


class WolkAboutProtocolMessageFactory(MessageFactory):
    """Serialize messages to be sent to WolkAbout IoT Platform."""

    DEVICE_TO_PLATFORM = "d2p/"
    CHANNEL_DELIMITER = "/"

    TIME = "time"

    PARAMETERS = "parameters"
    PULL_PARAMETERS = "pull_parameters"

    FEED_VALUES = "feed_values"
    PULL_FEED_VALUES = "pull_feed_values"

    FEED_REGISTRATION = "feed_registration"
    FEED_REMOVAL = "feed_removal"

    ATTRIBUTE_REGISTRATION = "attribute_registration"

    FILE_BINARY_REQUEST = "file_binary_request"
    FILE_LIST = "file_list"
    FILE_UPLOAD_STATUS = "file_upload_status"
    FILE_URL_DOWNLOAD_STATUS = "file_url_download_status"

    FIRMWARE_VERSION_UPDATE = "firmware_version_update"
    FIRMWARE_UPDATE_STATUS = "firmware_update_status"

    def __init__(self, device_key: str) -> None:
        """
        Create a factory for serializing messages.

        :param device_key: Device key to use when serializing messages
        :type device_key: str
        """
        self.device_key = device_key
        self.logger = logger_factory.logger_factory.get_logger(
            str(self.__class__.__name__)
        )
        self.logger.debug(f"Device key: {device_key}")
        self.common_topic = (
            self.DEVICE_TO_PLATFORM + self.device_key + self.CHANNEL_DELIMITER
        )

    def make_from_feed_value(
        self,
        reading: Union[Reading, List[Reading]],
        timestamp: Optional[int],
    ) -> Message:
        """
        Serialize feed value data.

        :param reading: Feed value data as (reference, value) or list of tuple
        :type reading: Union[Reading, List[Reading]]
        :param timestamp: Unix timestamp in ms. Default to current time if None
        :raises ValueError: Reading is invalid data type
        :returns: message
        :rtype: Message
        """
        topic = self.common_topic + self.FEED_VALUES

        if isinstance(reading, tuple):
            reference, value = reading
            feed_value = {reference: value}
        elif isinstance(reading, list):
            feed_value = dict(reading)
        else:
            raise ValueError(
                f"Expected reading as tuple or list, got {type(reading)}"
            )

        feed_value["timestamp"] = (
            timestamp if timestamp is not None else round(time() * 1000)
        )
        payload = [feed_value]

        message = Message(topic, json.dumps(payload))
        self.logger.debug(f"{message}")

        return message

    def make_time_request(self) -> Message:
        """
        Serialize message requesting platform timestamp.

        :returns: message
        :rtype: Message
        """
        topic = self.common_topic + self.TIME

        message = Message(topic)
        self.logger.debug(f"{message}")

        return message

    def make_pull_feed_values(self) -> Message:
        """
        Serialize message requesting any pending inbound feed values.

        :returns: message
        :rtype: Message
        """
        topic = self.common_topic + self.PULL_FEED_VALUES

        message = Message(topic)
        self.logger.debug(f"{message}")

        return message

    def make_from_parameters(
        self, parameters: Dict[str, Union[bool, int, float, str]]
    ) -> Message:
        """
        Serialize device parameters to be sent to the Platform.

        :param parameters: Device parameters
        :type parameters: Dict[str, Union[bool, int, float, str]]
        :returns: message
        :rtype: Message
        """
        topic = self.common_topic + self.PARAMETERS

        message = Message(topic, json.dumps(parameters))
        self.logger.debug(f"{message}")

        return message

    def make_pull_parameters(self) -> Message:
        """
        Serialize request to pull device parameters from the Platform.

        :returns: message
        :rtype: Message
        """
        topic = self.common_topic + self.PULL_PARAMETERS

        message = Message(topic)
        self.logger.debug(f"{message}")

        return message

    def make_feed_registration(
        self,
        name: str,
        reference: str,
        feed_type: FeedType,
        unit: Union[Unit, str],
    ) -> Message:
        """
        Serialize request to register a feed on the Platform.

        :param name: Feed name
        :type name: str
        :param reference: Unique identifier
        :type reference: str
        :param feed_type: Is the feed one or two-way communication
        :type feed_type: FeedType
        :param unit: Unit used to measure this feed
        :type unit: Union[Unit, str]
        :returns: message
        :rtype: Message
        """
        topic = self.common_topic + self.FEED_REGISTRATION

        payload = {
            "name": name,
            "reference": reference,
            "type": feed_type.value,
        }

        if isinstance(unit, Unit):
            payload["unitGuid"] = unit.value
        else:
            payload["unitGuid"] = unit

        message = Message(topic, json.dumps([payload]))
        self.logger.debug(f"{message}")

        return message

    def make_feed_removal(self, reference: str) -> Message:
        """
        Serialize request to remove a feed from the device on the Platform.

        :param reference: Unique identifier
        :type reference: str
        :returns: message
        :rtype: Message
        """
        topic = self.common_topic + self.FEED_REMOVAL

        payload = json.dumps([reference])

        message = Message(topic, payload)
        self.logger.debug(f"{message}")

        return message

    def make_attribute_registration(
        self, name: str, data_type: DataType, value: str
    ) -> Message:
        """
        Serialize request to register an attribute for the device.

        :param name: Unique identifier
        :type name: str
        :param data_type: Type of data this attribute holds
        :type data_type: DataType
        :param value: Value of the attribute
        :type value: str
        :returns: message
        :rtype: Message
        """
        topic = self.common_topic + self.ATTRIBUTE_REGISTRATION

        payload = [{"name": name, "dataType": data_type.value, "value": value}]

        message = Message(topic, json.dumps(payload))
        self.logger.debug(f"{message}")

        return message

    def make_from_package_request(
        self, file_name: str, chunk_index: int
    ) -> Message:
        """
        Request a package of the file from WolkAbout IoT Platform.

        :param file_name: Name of the file that contains the requested package
        :type file_name: str
        :param chunk_index: Index of the requested package
        :type chunk_index: int
        :returns: message
        :rtype: Message
        """
        self.logger.debug(
            f"file_name: '{file_name}', " f"chunk_index: {chunk_index}, "
        )
        topic = self.common_topic + self.FILE_BINARY_REQUEST

        payload = {
            "name": file_name,
            "chunkIndex": chunk_index,
        }
        message = Message(topic, json.dumps(payload))
        self.logger.debug(f"{message}")

        return message

    def make_from_file_list(
        self, file_list: List[Dict[str, Union[str, int]]]
    ) -> Message:
        """
        Serialize list of files present on device.

        :param file_list: Files present on device
        :type file_list: List[Dict[str, Union[str, int]]]
        :returns: message
        :rtype: Message
        """
        self.logger.debug(f"{file_list}")
        topic = self.common_topic + self.FILE_LIST

        message = Message(topic, json.dumps(file_list))
        self.logger.debug(f"{message}")

        return message

    def make_from_file_management_status(
        self, status: FileManagementStatus, file_name: str
    ) -> Message:
        """
        Serialize device's current file management status.

        :param status: Current file management status
        :type status: FileManagementStatus
        :param file_name: Name of file being transferred
        :type file_name: str
        :returns: message
        :rtype: Message
        """
        self.logger.debug(f" status: {status}, file_name: {file_name}")
        topic = self.common_topic + self.FILE_UPLOAD_STATUS

        payload = {"name": file_name, "status": status.status.value}
        if (
            status.status == FileManagementStatusType.ERROR
            and status.error is not None
        ):
            payload["error"] = status.error.value

        message = Message(topic, json.dumps(payload))
        self.logger.debug(f"{message}")

        return message

    def make_from_file_url_status(
        self,
        file_url: str,
        status: FileManagementStatus,
        file_name: Optional[str] = None,
    ) -> Message:
        """
        Serialize device's current file URL download status.

        :param file_url: URL from where the file is to be downloaded
        :type file_url: str
        :param status: Current file management status
        :type status: FileManagementStatus
        :param file_name: Only present when download of file is completed
        :type file_name: Optional[str]
        """
        self.logger.debug(
            f"file_url: {file_url}, status: {status}, file_name: {file_name}"
        )
        topic = self.common_topic + self.FILE_URL_DOWNLOAD_STATUS

        payload = {"fileUrl": file_url, "status": status.status.value}

        if (
            status.status == FileManagementStatusType.ERROR
            and status.error is not None
        ):
            payload["error"] = status.error.value

        if file_name is not None:
            payload["fileName"] = file_name

        message = Message(topic, json.dumps(payload))
        self.logger.debug(f"{message}")

        return message

    def make_from_firmware_update_status(
        self, firmware_update_status: FirmwareUpdateStatus
    ) -> Message:
        """
        Serialize firmware update status to be sent to WolkAbout IoT Platform.

        :param firmware_update_status: Firmware update status to be serialized
        :type firmware_status: FirmwareUpdateStatus
        :returns: message
        :rtype: Message
        """
        self.logger.debug(f"{firmware_update_status}")
        topic = self.common_topic + self.FIRMWARE_UPDATE_STATUS
        payload = {"status": firmware_update_status.status.value}

        if (
            firmware_update_status.status == FirmwareUpdateStatusType.ERROR
            and firmware_update_status.error is not None
        ):
            payload["error"] = firmware_update_status.error.value

        message = Message(topic, json.dumps(payload))
        self.logger.debug(f"{message}")

        return message



"""File transfer package model."""
#   Copyright 2020 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from dataclasses import dataclass


@dataclass
class FileTransferPackage:
    """
    Data from a file binary request response.

    :ivar previous_hash: Hash of the previous chunk
    :vartype previous_hash: bytes
    :ivar data: Requested chunk
    :vartype data: bytes
    :ivar current_hash: Hash of the current chunk
    :vartype current_hash: bytes
    """

    previous_hash: bytes
    data: bytes
    current_hash: bytes



"""Tests for WolkAboutProtocolMessageDeserializer."""
#   Copyright 2020 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import json
import logging
import sys
import unittest
from unittest.mock import MagicMock

sys.path.append("..")  # noqa

from wolk.model.device import Device
from wolk.model.message import Message
from wolk.model.file_transfer_package import FileTransferPackage
from wolk.wolkabout_protocol_message_deserializer import (
    WolkAboutProtocolMessageDeserializer as WAPMD,
)


class WolkAboutProtocolMessageDeserializerTests(unittest.TestCase):
    """Tests for deserializing messages using WolkAbout Protocol."""

    device = Device(
        key="some_key",
        password="some_password",
    )

    expected_topics = [
        WAPMD.PLATFORM_TO_DEVICE
        + device.key
        + WAPMD.CHANNEL_DELIMITER
        + WAPMD.TIME,
        WAPMD.PLATFORM_TO_DEVICE
        + device.key
        + WAPMD.CHANNEL_DELIMITER
        + WAPMD.FEED_VALUES,
        WAPMD.PLATFORM_TO_DEVICE
        + device.key
        + WAPMD.CHANNEL_DELIMITER
        + WAPMD.PARAMETERS,
        WAPMD.PLATFORM_TO_DEVICE
        + device.key
        + WAPMD.CHANNEL_DELIMITER
        + WAPMD.FILE_BINARY,
        WAPMD.PLATFORM_TO_DEVICE
        + device.key
        + WAPMD.CHANNEL_DELIMITER
        + WAPMD.FILE_DELETE,
        WAPMD.PLATFORM_TO_DEVICE
        + device.key
        + WAPMD.CHANNEL_DELIMITER
        + WAPMD.FILE_PURGE,
        WAPMD.PLATFORM_TO_DEVICE
        + device.key
        + WAPMD.CHANNEL_DELIMITER
        + WAPMD.FILE_LIST,
        WAPMD.PLATFORM_TO_DEVICE
        + device.key
        + WAPMD.CHANNEL_DELIMITER
        + WAPMD.FILE_UPLOAD_ABORT,
        WAPMD.PLATFORM_TO_DEVICE
        + device.key
        + WAPMD.CHANNEL_DELIMITER
        + WAPMD.FILE_UPLOAD_INITIATE,
        WAPMD.PLATFORM_TO_DEVICE
        + device.key
        + WAPMD.CHANNEL_DELIMITER
        + WAPMD.FILE_URL_ABORT,
        WAPMD.PLATFORM_TO_DEVICE
        + device.key
        + WAPMD.CHANNEL_DELIMITER
        + WAPMD.FILE_URL_INITIATE,
        WAPMD.PLATFORM_TO_DEVICE
        + device.key
        + WAPMD.CHANNEL_DELIMITER
        + WAPMD.FIRMWARE_ABORT,
        WAPMD.PLATFORM_TO_DEVICE
        + device.key
        + WAPMD.CHANNEL_DELIMITER
        + WAPMD.FIRMWARE_INSTALL,
    ]

    def setUp(self):
        """Set up commonly used values in tests."""
        self.maxDiff = None
        self.device = Device(
            key="some_key",
            password="some_password",
        )
        self.deserializer = WAPMD(self.device)

    def test_inbound_topics_match(self):
        """Test creating a deserializer and assert topic lists match."""
        self.assertEqual(
            set(self.expected_topics), set(self.deserializer.inbound_topics)
        )

    def test_get_inbound_topics(self):
        """Test inbound topics are valid for device."""
        self.assertEqual(
            self.expected_topics, self.deserializer.get_inbound_topics()
        )

    def test_is_time_response(self):
        """Test if message is keep alive response."""
        message = Message(self.deserializer.time_topic, None)

        self.assertTrue(self.deserializer.is_time_response(message))

    def test_is_firmware_install(self):
        """Test if message is firmware install command."""
        message = Message(self.deserializer.firmware_install_topic, None)

        self.assertTrue(self.deserializer.is_firmware_install(message))

    def test_is_firmware_abort(self):
        """Test if message is firmware abort command."""
        message = Message(self.deserializer.firmware_abort_topic, None)

        self.assertTrue(self.deserializer.is_firmware_abort(message))

    def test_is_file_binary(self):
        """Test if message is file binary."""
        message = Message(self.deserializer.file_binary_topic, None)

        self.assertTrue(self.deserializer.is_file_binary_response(message))

    def test_is_file_delete_command(self):
        """Test if message is file delete command."""
        message = Message(self.deserializer.file_delete_topic, None)

        self.assertTrue(self.deserializer.is_file_delete_command(message))

    def test_is_file_purge_command(self):
        """Test if message is file purge command."""
        message = Message(self.deserializer.file_purge_topic, None)

        self.assertTrue(self.deserializer.is_file_purge_command(message))

    def test_is_file_upload_initiate(self):
        """Test if message is file upload initiate."""
        message = Message(self.deserializer.file_upload_initiate_topic, None)

        self.assertTrue(self.deserializer.is_file_upload_initiate(message))

    def test_is_file_upload_abort(self):
        """Test if message is file upload abort."""
        message = Message(self.deserializer.file_upload_abort_topic, None)

        self.assertTrue(self.deserializer.is_file_upload_abort(message))

    def test_is_file_url_initiate(self):
        """Test if message is file URL initiate."""
        message = Message(self.deserializer.file_url_initiate_topic, None)

        self.assertTrue(self.deserializer.is_file_url_initiate(message))

    def test_is_file_url_abort(self):
        """Test if message is file URL abort."""
        message = Message(self.deserializer.file_url_abort_topic, None)

        self.assertTrue(self.deserializer.is_file_url_abort(message))

    def test_parse_time_response(self):
        """Test parse keep alive response message."""
        self.deserializer.logger.setLevel(logging.CRITICAL)
        timestamp = 123

        incoming_topic = self.deserializer.time_topic
        incoming_payload = bytearray(json.dumps(timestamp), "utf-8")
        incoming_message = Message(incoming_topic, incoming_payload)

        expected = timestamp

        self.assertEqual(
            expected, self.deserializer.parse_time_response(incoming_message)
        )

    def test_parse_firmware_install(self):
        """Test parse firmware install command."""
        self.deserializer.logger.setLevel(logging.CRITICAL)
        file_name = "install_me.bin"

        incoming_topic = self.deserializer.firmware_install_topic
        incoming_payload = bytearray(json.dumps(file_name), "utf-8")
        incoming_message = Message(incoming_topic, incoming_payload)

        expected = file_name

        self.assertEqual(
            expected,
            self.deserializer.parse_firmware_install(incoming_message),
        )

    def test_parse_firmware_install_invalid(self):
        """Test parse invalid firmware install command."""
        self.deserializer.logger.setLevel(logging.CRITICAL)

        incoming_topic = self.deserializer.firmware_install_topic
        incoming_payload = bytearray('{"file_name": file_name', "utf-8")
        incoming_message = Message(incoming_topic, incoming_payload)

        expected = ""

        self.assertEqual(
            expected,
            self.deserializer.parse_firmware_install(incoming_message),
        )

    def test_parse_file_binary_invalid(self):
        """Test parse invalid file binary payload."""
        self.deserializer.logger.setLevel(logging.CRITICAL)
        file_name = "install_me.bin"

        incoming_topic = self.deserializer.file_binary_topic
        incoming_payload = bytearray(
            json.dumps({"file_name": file_name}), "utf-8"
        )
        incoming_message = Message(incoming_topic, incoming_payload)

        expected = FileTransferPackage(b"", b"", b"")

        self.assertEqual(
            expected, self.deserializer.parse_file_binary(incoming_message)
        )

    def test_parse_file_binary_invalid_size(self):
        """Test parse invalid file binary payload size."""
        self.deserializer.logger.setLevel(logging.CRITICAL)
        previous_hash = 15 * b"\x00"
        data = b"\x00"
        current_hash = 15 * b"\x00"

        incoming_topic = self.deserializer.file_binary_topic
        incoming_payload = previous_hash + data + current_hash
        incoming_message = Message(incoming_topic, incoming_payload)

        expected = FileTransferPackage(b"", b"", b"")

        self.assertEqual(
            expected, self.deserializer.parse_file_binary(incoming_message)
        )

    def test_parse_file_binary(self):
        """Test parse file binary."""
        self.deserializer.logger.setLevel(logging.CRITICAL)
        previous_hash = 32 * b"\x00"
        data = 32 * b"\x00"
        current_hash = 32 * b"\x00"

        incoming_topic = self.deserializer.file_binary_topic
        incoming_payload = previous_hash + data + current_hash
        incoming_message = Message(incoming_topic, incoming_payload)

        expected = FileTransferPackage(previous_hash, data, current_hash)

        self.assertEqual(
            expected, self.deserializer.parse_file_binary(incoming_message)
        )

    def test_parse_file_delete_command(self):
        """Test parse file delete command."""
        self.deserializer.logger.setLevel(logging.CRITICAL)
        file_name = "delete_me.bin"
        expected = [file_name]

        incoming_topic = self.deserializer.file_delete_topic
        incoming_payload = bytearray(json.dumps([file_name]), "utf-8")
        incoming_message = Message(incoming_topic, incoming_payload)

        self.assertEqual(
            expected,
            self.deserializer.parse_file_delete_command(incoming_message),
        )

    def test_parse_file_delete_command_invalid(self):
        """Test parse file delete invalid command."""
        self.deserializer.logger.setLevel(logging.CRITICAL)
        expected = []

        incoming_topic = self.deserializer.file_delete_topic
        incoming_payload = bytearray('{"file_name": file_name', "utf-8")
        incoming_message = Message(incoming_topic, incoming_payload)

        self.assertEqual(
            expected,
            self.deserializer.parse_file_delete_command(incoming_message),
        )

    def test_parse_file_url(self):
        """Test parse file URL command."""
        self.deserializer.logger.setLevel(logging.CRITICAL)
        file_url = "http://hello.there.hi/resource.png"
        expected = file_url

        incoming_topic = self.deserializer.file_url_initiate_topic
        incoming_payload = bytearray(json.dumps(file_url), "utf-8")
        incoming_message = Message(incoming_topic, incoming_payload)

        self.assertEqual(
            expected, self.deserializer.parse_file_url(incoming_message)
        )

    def test_parse_file_url_invalid(self):
        """Test parse file URL invalid command."""
        self.deserializer.logger.setLevel(logging.CRITICAL)
        expected = ""

        incoming_topic = self.deserializer.file_url_initiate_topic
        incoming_payload = bytearray('{"file_url": file_url', "utf-8")
        incoming_message = Message(incoming_topic, incoming_payload)

        self.assertEqual(
            expected, self.deserializer.parse_file_url(incoming_message)
        )

    def test_parse_file_initiate(self):
        """Test parse file initiate command."""
        self.deserializer.logger.setLevel(logging.CRITICAL)
        file_name = "file.bin"
        file_size = 128
        file_hash = "some_hash"
        expected = (file_name, file_size, file_hash)

        incoming_topic = self.deserializer.file_upload_initiate_topic
        incoming_payload = bytearray(
            json.dumps(
                {
                    "name": file_name,
                    "size": file_size,
                    "hash": file_hash,
                }
            ),
            "utf-8",
        )
        incoming_message = Message(incoming_topic, incoming_payload)

        self.assertEqual(
            expected, self.deserializer.parse_file_initiate(incoming_message)
        )

    def test_parse_file_initiate_invalid(self):
        """Test parse file initiate invalid command."""
        self.deserializer.logger.setLevel(logging.CRITICAL)
        file_name = "file.bin"
        file_size = 128
        file_hash = "some_hash"
        expected = ("", 0, "")

        incoming_topic = self.deserializer.file_upload_initiate_topic
        incoming_payload = bytearray(
            json.dumps(
                {
                    "obviously_wrong_name": file_name,
                    "size": file_size,
                    "hash": file_hash,
                }
            ),
            "utf-8",
        )
        incoming_message = Message(incoming_topic, incoming_payload)

        self.assertEqual(
            expected, self.deserializer.parse_file_initiate(incoming_message)
        )

    def test_parse_parameters(self):
        """Test parsing the parameters message received from the Platform."""
        self.deserializer.logger.setLevel(logging.CRITICAL)
        parameters = {"FIRMWARE_UPDATE_CHECK_TIME": 1}

        expected = parameters

        incoming_topic = self.deserializer.parameters_topic
        incoming_payload = json.dumps(parameters)

        incoming_message = Message(incoming_topic, incoming_payload)

        self.assertEqual(
            expected, self.deserializer.parse_parameters(incoming_message)
        )

    def test_parse_parameters_exception(self):
        """Test parsing faulty parameters message from the Platform."""
        self.deserializer.logger.setLevel(logging.CRITICAL)
        self.deserializer.logger.exception = MagicMock()

        expected = {}

        incoming_topic = self.deserializer.parameters_topic
        incoming_payload = "{"

        incoming_message = Message(incoming_topic, incoming_payload)

        self.assertEqual(
            expected, self.deserializer.parse_parameters(incoming_message)
        )
        self.deserializer.logger.exception.assert_called_once()

    def test_parse_feed_values(self):
        """Test parsing the feed values message received from the Platform."""
        self.deserializer.logger.setLevel(logging.CRITICAL)
        feed_values = [{"reference": True, "timestamp": 123}]

        expected = feed_values

        incoming_topic = self.deserializer.feed_values_topic
        incoming_payload = json.dumps(feed_values)

        incoming_message = Message(incoming_topic, incoming_payload)

        self.assertEqual(
            expected, self.deserializer.parse_feed_values(incoming_message)
        )

    def test_parse_feed_values_exception(self):
        """Test parsing faulty feed values message from the Platform."""
        self.deserializer.logger.setLevel(logging.CRITICAL)
        self.deserializer.logger.exception = MagicMock()

        expected = []

        incoming_topic = self.deserializer.feed_values_topic
        incoming_payload = "["

        incoming_message = Message(incoming_topic, incoming_payload)

        self.assertEqual(
            expected, self.deserializer.parse_feed_values(incoming_message)
        )
        self.deserializer.logger.exception.assert_called_once()



"""Message storage implemented via double ended queue."""
#   Copyright 2020 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from collections import deque
from typing import Optional

from wolk import logger_factory
from wolk.interface.message_queue import MessageQueue
from wolk.model.message import Message


class MessageDeque(MessageQueue):
    """
    Store messages before they are sent to the WolkAbout IoT Platform.

    :ivar logger: Logger instance issued by wolk.LoggerFactory
    :vartype logger: logging.Logger
    :ivar queue: Double ended queue used to store messages
    :vartype queue: collections.deque
    """

    def __init__(self) -> None:
        """Create a double ended queue to store messages."""
        self.queue: deque = deque()
        self.logger = logger_factory.logger_factory.get_logger(
            str(self.__class__.__name__)
        )

    def put(self, message: Message) -> bool:
        """
        Add the message to the queue.

        :param message: Message to place in the queue
        :type message: Message
        :returns: success
        :rtype: bool
        """
        if not message:
            self.logger.error("Nothing to store!")
            return False

        self.queue.append(message)
        self.logger.debug(
            f"Stored message: {message} - Queue size: {len(self.queue)}"
        )
        return True

    def get(self) -> Optional[Message]:
        """
        Take the first message from the queue.

        :returns: message
        :rtype: Optional[Message]
        """
        if len(self.queue) == 0:
            return None
        message = self.queue.popleft()
        self.logger.debug(
            f"Returning message: {message} " f"- Queue size: {len(self.queue)}"
        )
        return message

    def peek(self) -> Optional[Message]:
        """
        Return the first message from the queue without removing it.

        :returns: message
        :rtype: Optional[Message]
        """
        if len(self.queue) == 0:
            self.logger.debug("Empty queue")
            return None

        message = self.queue[0]
        self.logger.debug(
            f"Returning message: {message} " f"- Queue size: {len(self.queue)}"
        )
        return message



"""Tests for WolkConnect."""
#   Copyright 2020 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import logging
import os
import sys
import unittest
from unittest.mock import MagicMock

sys.path.append("..")

from wolk.wolk_connect import WolkConnect
from wolk.model.device import Device
from wolk.model.data_delivery import DataDelivery
from wolk.model.data_type import DataType
from wolk.model.feed_type import FeedType
from wolk.model.unit import Unit
from wolk.model.message import Message
from wolk.model.firmware_update_status import (
    FirmwareUpdateStatus,
    FirmwareUpdateStatusType,
)
from wolk.model.file_management_status import (
    FileManagementStatus,
    FileManagementStatusType,
)
from wolk.interface.firmware_handler import FirmwareHandler
from wolk.message_deque import MessageDeque
from wolk.wolkabout_protocol_message_deserializer import (
    WolkAboutProtocolMessageDeserializer as WAPMD,
)
from wolk.wolkabout_protocol_message_factory import (
    WolkAboutProtocolMessageFactory as WAPMF,
)
from wolk.mqtt_connectivity_service import MQTTConnectivityService


class TestWolkConnect(unittest.TestCase):
    """Tests for WolkConnect class."""

    class MockFirmwareHandler(FirmwareHandler):
        """Mock firmware installer class that whose methods will be mocked."""

        def install_firmware(self, firmware_file_path: str) -> None:
            """
            Handle the installation of the firmware file.

            :param firmware_file_path: Path where the firmware file is located
            :type firmware_file_path: str
            """
            raise NotImplementedError()

        def get_current_version(self) -> str:
            """
            Obtain device's current firmware version.

            :returns: version
            :rtype: str
            """
            raise NotImplementedError()

    def setUp(self) -> None:
        """Set up values that are commonly used in tests."""
        self.maxDiff = None
        self.device_key = "some_key"
        self.device_password = "some_password"
        self.device = Device(self.device_key, self.device_password)

        self.wolk_device = WolkConnect(self.device)
        self.wolk_device.logger.setLevel(logging.CRITICAL)
        self.file_directory = "test_files"

        self.firmware_handler = self.MockFirmwareHandler()

        self.message = Message("some_topic", "payload")

        self.file_name = "file"
        self.file_url = "file_url"

    def test_init_default_server(self):
        """Test creating instance with default server parameters."""
        self.assertIsNotNone(self.wolk_device.connectivity_service.ca_cert)

    def test_init_custom_server_unsecure(self):
        """Test creating instance with server on unsecure port."""
        self.wolk_device = WolkConnect(self.device, "some_host", 1883)
        self.assertIsNone(self.wolk_device.connectivity_service.ca_cert)

    def test_init_custom_server_secure(self):
        """Test creating instance with server on secure port."""
        self.wolk_device = WolkConnect(
            self.device, "some_host", 1883, "some_cert"
        )
        self.assertIsNotNone(self.wolk_device.connectivity_service.ca_cert)

    def test_with_feed_value_handler_not_callable(self):
        """Test adding faulty feed value handler that is not callable."""
        self.assertRaises(
            ValueError,
            self.wolk_device.with_incoming_feed_value_handler,
            None,
        )

    def test_with_feed_handler_invalid_signature(self):
        """Test adding faulty feed value handler with bad signature."""
        self.assertRaises(
            ValueError,
            self.wolk_device.with_incoming_feed_value_handler,
            lambda a, b, c: a,
        )

    def test_with_configuration_valid(self):
        """Test adding configurations with valid handler and provider."""
        self.wolk_device.with_incoming_feed_value_handler(lambda a: None)

        self.assertIsNotNone(self.wolk_device.incoming_feed_value_handler)

    def test_with_file_management(self):
        """Test enabling file management module."""
        self.wolk_device.with_file_management(self.file_directory, 1024)
        self.assertTrue(os.path.exists(self.file_directory))
        os.rmdir(self.file_directory)

    def test_with_file_management_with_custom_url_download(self):
        """Test enabling file management module with custom URL download."""

        def _downloader(a, b):
            pass

        self.wolk_device.with_file_management(
            self.file_directory, 1024, _downloader
        )
        os.rmdir(self.file_directory)
        self.assertEqual(
            _downloader, self.wolk_device.file_management.url_downloader
        )

    def test_with_firmware_update_no_file_management(self):
        """Test enabling firmware update module fails if no file management."""
        self.assertRaises(
            RuntimeError, self.wolk_device.with_firmware_update, 12
        )

    def test_with_firmware_update_and_file_management(self):
        """Test enabling firmware update module with file management module."""
        self.wolk_device.with_file_management(self.file_directory, 1024)
        os.rmdir(self.file_directory)
        self.firmware_handler.get_current_version = MagicMock(
            return_value="1.0"
        )

        self.wolk_device.with_firmware_update(self.firmware_handler)

        self.assertIsNotNone(self.wolk_device.file_management)
        self.assertIsNotNone(self.wolk_device.firmware_update)

    def test_with_custom_message_queue_invalid_instance(self):
        """Test using custom message queue with passing bad message queue."""
        self.assertRaises(
            ValueError, self.wolk_device.with_custom_message_queue, 1
        )

    def test_with_custom_message_queue_valid_instance(self):
        """Test using custom message queue with passing good message queue."""

        class MockMessageQueue(MessageDeque):
            pass

        self.wolk_device.with_custom_message_queue(MockMessageQueue())

        self.assertIsInstance(self.wolk_device.message_queue, MockMessageQueue)

    def test_with_custom_protocol_bad_factory(self):
        """Test using custom protocol with bad message factory."""
        self.assertRaises(
            ValueError, self.wolk_device.with_custom_protocol, 12, 34
        )

    def test_with_custom_protocol_bad_deserializer(self):
        """Test using custom protocol with bad message deserializer."""

        class MockFactory(WAPMF):
            pass

        self.assertRaises(
            ValueError,
            self.wolk_device.with_custom_protocol,
            MockFactory(self.device_key),
            34,
        )

    def test_with_custom_protocol_valid(self):
        """Test using custom protocol with valid factory and deserializer."""

        class MockFactory(WAPMF):
            pass

        class MockDeserializer(WAPMD):
            pass

        self.wolk_device.with_custom_protocol(
            MockFactory(self.device_key), MockDeserializer(self.device)
        )

        self.assertIsInstance(self.wolk_device.message_factory, MockFactory)
        self.assertIsInstance(
            self.wolk_device.message_deserializer, MockDeserializer
        )

    def test_with_custom_connectivity_invalid_instance(self):
        """Test using custom connectivity with bad instance."""
        self.assertRaises(
            ValueError, self.wolk_device.with_custom_connectivity, 1
        )

    def test_with_custom_connectivity_valid_instance(self):
        """Test using custom connectivity with valid instance."""

        class MockCS(MQTTConnectivityService):
            pass

        mock_cs = MockCS(self.device, [])
        mock_cs.set_inbound_message_listener = MagicMock()

        self.wolk_device.with_custom_connectivity(mock_cs)

        mock_cs.set_inbound_message_listener.assert_called_once_with(
            self.wolk_device._on_inbound_message
        )

    def test_connect_already_connected(self):
        """Test connecting when already connected."""
        self.wolk_device.connectivity_service.is_connected = MagicMock(
            return_value=True
        )
        self.wolk_device.logger.info = MagicMock()

        self.wolk_device.connect()
        self.wolk_device.logger.info.assert_called_once()

    def test_connect_cs_raises_exception(self):
        """Test connecting with connectivity service failing."""
        self.wolk_device.connectivity_service.connect = MagicMock(
            side_effect=Exception()
        )
        self.wolk_device.logger.exception = MagicMock()

        self.wolk_device.connect()
        self.wolk_device.logger.exception.assert_called_once()

    def test_connect_fail_to_connect(self):
        """Test connecting fails."""
        self.wolk_device.connectivity_service.is_connected = MagicMock(
            return_value=False
        )
        self.wolk_device.connectivity_service.connect = MagicMock(
            return_value=False
        )
        self.wolk_device.logger.exception = MagicMock()

        self.wolk_device.connect()
        self.assertEqual(
            2, self.wolk_device.connectivity_service.is_connected.call_count
        )

    def test_connect_connects(self):
        """Test connecting passes."""
        self.wolk_device.connectivity_service.is_connected = MagicMock()
        self.wolk_device.connectivity_service.is_connected.side_effect = [
            False,
            False,
            True,
        ]
        self.wolk_device.connectivity_service.connect = MagicMock(
            return_value=True
        )
        self.wolk_device.connect()
        self.assertEqual(
            2, self.wolk_device.connectivity_service.is_connected.call_count
        )

    def test_connect_publish_file_list_fails(self):
        """Test connecting passes and fails to publishes file list."""
        self.wolk_device.with_file_management(self.file_directory, 1024)
        self.wolk_device.file_management.get_file_list = MagicMock(
            return_value=[]
        )
        self.wolk_device.connectivity_service.publish = MagicMock(
            return_value=False
        )
        self.wolk_device.message_queue.put = MagicMock()

        self.wolk_device.connectivity_service.is_connected = MagicMock()
        self.wolk_device.connectivity_service.is_connected.side_effect = [
            False,
            True,
        ]
        self.wolk_device.connectivity_service.connect = MagicMock(
            return_value=True
        )
        self.wolk_device.connectivity_service.publish = MagicMock(
            return_value=False
        )
        self.wolk_device.connect()
        self.wolk_device.message_queue.put.assert_any_call(
            Message("d2p/some_key/file_list", "[]")
        )
        os.rmdir(self.file_directory)

    def test_connect_publish_file_list(self):
        """Test connecting passes and publishes file list."""
        self.wolk_device.with_file_management(self.file_directory, 1024)
        self.wolk_device.file_management.get_file_list = MagicMock(
            return_value=[]
        )
        self.wolk_device.connectivity_service.publish = MagicMock(
            return_value=True
        )
        self.wolk_device.message_queue.put = MagicMock()

        self.wolk_device.connectivity_service.is_connected = MagicMock()
        self.wolk_device.connectivity_service.is_connected.side_effect = [
            False,
            True,
        ]
        self.wolk_device.connectivity_service.connect = MagicMock(
            return_value=True
        )
        self.wolk_device.connect()
        self.wolk_device.message_queue.put.assert_not_called()
        os.rmdir(self.file_directory)

    def test_connect_publish_firmware_version_fails(self):
        """Test connecting passes and firmware version fails to send."""
        self.firmware_handler.get_current_version = MagicMock(
            return_value="1.0"
        )

        self.wolk_device.with_file_management(self.file_directory, 1024)
        self.wolk_device.with_firmware_update(self.firmware_handler)
        self.wolk_device.file_management = None
        self.wolk_device.connectivity_service.publish = MagicMock(
            return_value=False
        )
        self.wolk_device.message_queue.put = MagicMock()

        self.wolk_device.connectivity_service.is_connected = MagicMock()
        self.wolk_device.connectivity_service.is_connected.side_effect = [
            False,
            True,
        ]
        self.wolk_device.connectivity_service.connect = MagicMock(
            return_value=True
        )
        self.wolk_device.connect()
        self.wolk_device.message_queue.put.assert_called_once()
        os.rmdir(self.file_directory)

    def test_connect_publish_firmware_version_passes(self):
        """Test connecting passes and firmware version is sent."""
        self.firmware_handler.get_current_version = MagicMock(
            return_value="1.0"
        )

        self.wolk_device.with_file_management(self.file_directory, 1024)
        self.wolk_device.with_firmware_update(self.firmware_handler)
        self.wolk_device.file_management = None
        self.wolk_device.connectivity_service.publish = MagicMock(
            return_value=True
        )
        self.wolk_device.message_queue.put = MagicMock()

        self.wolk_device.connectivity_service.is_connected = MagicMock()
        self.wolk_device.connectivity_service.is_connected.side_effect = [
            False,
            True,
        ]
        self.wolk_device.connectivity_service.connect = MagicMock(
            return_value=True
        )
        self.wolk_device.connect()
        self.wolk_device.message_queue.put.assert_not_called()
        os.rmdir(self.file_directory)

    def test_connect_publish_pull_device(self):
        """Test connecting passes and PULL type device calls pull functions."""
        self.wolk_device.device.data_delivery = DataDelivery.PULL
        self.wolk_device.connectivity_service.publish = MagicMock(
            return_value=True
        )
        self.wolk_device.pull_parameters = MagicMock()
        self.wolk_device.pull_feed_values = MagicMock()

        self.wolk_device.connectivity_service.is_connected = MagicMock()
        self.wolk_device.connectivity_service.is_connected.side_effect = [
            False,
            True,
        ]
        self.wolk_device.connectivity_service.connect = MagicMock(
            return_value=True
        )
        self.wolk_device.connect()
        self.wolk_device.pull_parameters.assert_called_once()
        self.wolk_device.pull_feed_values.assert_called_once()

    def test_disconnect_not_connected(self):
        """Test calling disconnect when not connected."""
        self.wolk_device.logger.setLevel(logging.DEBUG)
        self.wolk_device.logger.debug = MagicMock()
        self.wolk_device.connectivity_service.is_connected = MagicMock(
            return_value=False
        )
        self.wolk_device.disconnect()
        self.wolk_device.logger.debug.assert_not_called()

    def test_disconnect_when_connected(self):
        """Test calling disconnect when connected."""
        self.wolk_device.logger.setLevel(logging.DEBUG)
        self.wolk_device.logger.debug = MagicMock()
        self.wolk_device.connectivity_service.is_connected = MagicMock(
            return_value=True
        )
        self.wolk_device.connectivity_service.disconnect = MagicMock()
        self.wolk_device.disconnect()
        self.wolk_device.logger.debug.assert_called_once()

    def test_publish_not_connected(self):
        """Test publishing when not connected."""
        self.wolk_device.logger.setLevel(logging.WARNING)
        self.wolk_device.logger.warning = MagicMock()
        self.wolk_device.connectivity_service.is_connected = MagicMock(
            return_value=False
        )
        self.wolk_device.publish()
        self.wolk_device.logger.warning.assert_called_once()

    def test_publish_emtpy_queue(self):
        """Test publishing when queue is empty."""
        self.wolk_device.connectivity_service.is_connected = MagicMock(
            return_value=True
        )
        self.wolk_device.message_queue.peek = MagicMock(return_value=None)
        self.wolk_device.publish()
        self.wolk_device.message_queue.peek.assert_called_once()

    def test_publish_fail_to_publish(self):
        """Test publishing and failing to publish message."""
        self.wolk_device.logger.setLevel(logging.WARNING)
        self.wolk_device.logger.warning = MagicMock()
        self.wolk_device.connectivity_service.is_connected = MagicMock(
            return_value=True
        )
        self.wolk_device.message_queue.peek = MagicMock(return_value=True)
        self.wolk_device.connectivity_service.publish = MagicMock(
            return_value=False
        )
        self.wolk_device.publish()
        self.wolk_device.logger.warning.assert_called_once()

    def test_publish_success(self):
        """Test publishing successfully."""
        self.wolk_device.logger.setLevel(logging.WARNING)
        self.wolk_device.logger.warning = MagicMock()
        self.wolk_device.connectivity_service.is_connected = MagicMock(
            return_value=True
        )
        self.wolk_device.message_queue.peek = MagicMock()
        self.wolk_device.message_queue.peek.side_effect = [True, None]
        self.wolk_device.connectivity_service.publish = MagicMock(
            return_value=True
        )
        self.wolk_device.message_queue.get = MagicMock()
        self.wolk_device.publish()
        self.wolk_device.message_queue.get.assert_called_once()

    def test_on_inbound_message_binary_topic(self):
        """Test on inbound message with 'binary' in topic."""
        self.wolk_device.logger.setLevel(logging.WARNING)
        message = Message("binary", "payload")
        self.wolk_device.logger.warning = MagicMock()

        self.wolk_device._on_inbound_message(message)
        self.wolk_device.logger.warning.assert_called_once()

    def test_on_inbound_message_unknown(self):
        """Test on inbound message for unknown message."""
        self.wolk_device.logger.setLevel(logging.WARNING)
        self.wolk_device.logger.warning = MagicMock()

        self.wolk_device._on_inbound_message(self.message)
        self.wolk_device.logger.warning.assert_called_once()

    def test_on_inbound_message_actuation_no_handlers(self):
        """Test on inbound actuation message but no actuation handler set."""
        self.wolk_device.logger.setLevel(logging.WARNING)
        self.wolk_device.logger.warning = MagicMock()
        self.wolk_device.message_deserializer.is_actuation_command = MagicMock(
            return_value=True
        )

        self.wolk_device._on_inbound_message(self.message)
        self.wolk_device.logger.warning.assert_called_once()

    def test_on_inbound_message_time_response(self):
        """Test on inbound time response message."""
        timestamp = 1

        self.wolk_device.message_deserializer.is_time_response = MagicMock(
            return_value=True
        )
        self.wolk_device.message_deserializer.parse_time_response = MagicMock(
            return_value=timestamp
        )
        self.wolk_device._on_inbound_message(self.message)
        self.assertEqual(timestamp, self.wolk_device.last_platform_timestamp)

    def test_on_inbound_message_file_management_message(self):
        """Test on inbound file management message."""
        self.wolk_device._on_file_management_message = MagicMock()
        self.wolk_device.message_deserializer.is_file_management_message = (
            MagicMock(return_value=True)
        )
        self.wolk_device._on_inbound_message(self.message)
        self.wolk_device._on_file_management_message.assert_called_once()

    def test_on_inbound_message_firmware_message(self):
        """Test on inbound firmware message."""
        self.wolk_device._on_firmware_message = MagicMock()
        self.wolk_device.message_deserializer.is_firmware_message = MagicMock(
            return_value=True
        )
        self.wolk_device._on_inbound_message(self.message)
        self.wolk_device._on_firmware_message.assert_called_once()

    def test_on_file_management_message_no_module_fail_to_send(self):
        """Test on file management message with no module."""
        self.wolk_device.message_factory.make_from_file_management_status = (
            MagicMock()
        )
        self.wolk_device.connectivity_service.publish = MagicMock(
            return_value=False
        )
        self.wolk_device.message_queue.put = MagicMock()

        self.wolk_device._on_file_management_message(self.message)
        self.wolk_device.message_queue.put.assert_called_once()

    def test_on_file_management_message_no_module_sends_error(self):
        """Test on file management message with no module, sends message."""
        self.wolk_device.message_factory.make_from_file_management_status = (
            MagicMock()
        )
        self.wolk_device.connectivity_service.publish = MagicMock(
            return_value=True
        )
        self.wolk_device.message_queue.put = MagicMock()

        self.wolk_device._on_file_management_message(self.message)
        self.wolk_device.message_queue.put.assert_not_called()

    def test_on_file_management_message_invalid_file_upload_init(self):
        """Test on file management message - invalid file upload initiate."""
        self.wolk_device.with_file_management(self.file_directory, 1024)
        os.rmdir(self.file_directory)

        self.wolk_device.message_deserializer.is_file_upload_initiate = (
            MagicMock(return_value=True)
        )
        self.wolk_device.message_deserializer.parse_file_initiate = MagicMock(
            return_value=("", 0, b"")
        )
        self.wolk_device.file_management.handle_upload_initiation = MagicMock()
        self.wolk_device._on_file_management_message(self.message)
        self.wolk_device.file_management.handle_upload_initiation.assert_not_called()

    def test_on_file_management_message_file_upload_init(self):
        """Test on file management message file upload initiate."""
        self.wolk_device.with_file_management(self.file_directory, 1024)
        os.rmdir(self.file_directory)

        self.wolk_device.message_deserializer.is_file_upload_initiate = (
            MagicMock(return_value=True)
        )
        self.wolk_device.message_deserializer.parse_file_initiate = MagicMock(
            return_value=("file", 0, b"")
        )
        self.wolk_device.file_management.handle_upload_initiation = MagicMock()
        self.wolk_device._on_file_management_message(self.message)
        self.wolk_device.file_management.handle_upload_initiation.assert_called_once_with(
            "file", 0, b""
        )

    def test_on_file_management_message_file_binary_response(self):
        """Test on file management message file binary response."""
        self.wolk_device.with_file_management(self.file_directory, 1024)
        os.rmdir(self.file_directory)

        self.wolk_device.message_deserializer.is_file_binary_response = (
            MagicMock(return_value=True)
        )
        self.wolk_device.message_deserializer.parse_file_binary = MagicMock(
            return_value=True
        )
        self.wolk_device.file_management.handle_file_binary_response = (
            MagicMock()
        )
        self.wolk_device._on_file_management_message(self.message)
        self.wolk_device.file_management.handle_file_binary_response.assert_called_once_with(
            True
        )

    def test_on_file_management_message_file_upload_abort(self):
        """Test on file management message file upload abort."""
        self.wolk_device.with_file_management(self.file_directory, 1024)
        os.rmdir(self.file_directory)

        self.wolk_device.message_deserializer.is_file_upload_abort = MagicMock(
            return_value=True
        )
        self.wolk_device.file_management.handle_file_upload_abort = MagicMock()
        self.wolk_device._on_file_management_message(self.message)
        self.wolk_device.file_management.handle_file_upload_abort.assert_called_once()

    def test_on_file_management_message_file_url_abort(self):
        """Test on file management message file URL abort."""
        self.wolk_device.with_file_management(self.file_directory, 1024)
        os.rmdir(self.file_directory)

        self.wolk_device.message_deserializer.is_file_url_abort = MagicMock(
            return_value=True
        )
        self.wolk_device.file_management.handle_file_upload_abort = MagicMock()
        self.wolk_device._on_file_management_message(self.message)
        self.wolk_device.file_management.handle_file_upload_abort.assert_called_once()

    def test_on_file_management_message_invalid_file_url_init(self):
        """Test on file management message - invalid file URL initiate."""
        self.wolk_device.with_file_management(self.file_directory, 1024)
        os.rmdir(self.file_directory)

        self.wolk_device.message_deserializer.is_file_url_initiate = MagicMock(
            return_value=True
        )
        self.wolk_device.message_deserializer.parse_file_url = MagicMock(
            return_value=""
        )
        self.wolk_device.file_management.handle_url_initiation = MagicMock()
        self.wolk_device._on_file_management_message(self.message)
        self.wolk_device.file_management.handle_url_initiation.assert_not_called()

    def test_on_file_management_message_file_url_init(self):
        """Test on file management message file URL initiate."""
        self.wolk_device.with_file_management(self.file_directory, 1024)
        os.rmdir(self.file_directory)

        self.wolk_device.message_deserializer.is_file_url_initiate = MagicMock(
            return_value=True
        )
        self.wolk_device.message_deserializer.parse_file_url = MagicMock(
            return_value="URL"
        )
        self.wolk_device.file_management.handle_file_url_download_initiation = (
            MagicMock()
        )
        self.wolk_device._on_file_management_message(self.message)
        self.wolk_device.file_management.handle_file_url_download_initiation.assert_called_once_with(
            "URL"
        )

    def test_on_file_management_message_file_file_list_(self):
        """Test on file management message file URL initiate."""
        self.wolk_device.with_file_management(self.file_directory, 1024)
        os.rmdir(self.file_directory)

        self.wolk_device.message_deserializer.is_file_url_initiate = MagicMock(
            return_value=True
        )
        self.wolk_device.message_deserializer.parse_file_url = MagicMock(
            return_value="URL"
        )
        self.wolk_device.file_management.handle_file_url_download_initiation = (
            MagicMock()
        )
        self.wolk_device._on_file_management_message(self.message)
        self.wolk_device.file_management.handle_file_url_download_initiation.assert_called_once_with(
            "URL"
        )

    def test_on_file_management_message_file_list_request_publishes(self):
        """Test on file list request fails to publish and puts in queue."""
        self.wolk_device.with_file_management(self.file_directory, 1024)
        os.rmdir(self.file_directory)

        self.wolk_device.message_deserializer.is_file_list_request = MagicMock(
            return_value=True
        )
        self.wolk_device.file_management.get_file_list = MagicMock(
            return_value=[]
        )
        self.wolk_device.connectivity_service.publish = MagicMock(
            return_value=True
        )
        self.wolk_device.message_queue.put = MagicMock()

        self.wolk_device._on_file_management_message(self.message)

        self.wolk_device.message_queue.put.assert_not_called()

    def test_on_file_management_message_file_delete_invalid_name(self):
        """Test receiving invalid file delete command."""
        self.wolk_device.with_file_management(self.file_directory, 1024)
        os.rmdir(self.file_directory)

        self.wolk_device.message_deserializer.is_file_delete_command = (
            MagicMock(return_value=True)
        )
        self.wolk_device.message_deserializer.parse_file_delete_command = (
            MagicMock(return_value="")
        )
        self.wolk_device.file_management.handle_file_delete = MagicMock()

        self.wolk_device._on_file_management_message(self.message)

        self.wolk_device.file_management.handle_file_delete.assert_not_called()

    def test_on_file_management_message_file_delete_fail_to_publish(self):
        """Test receiving file delete command and fail to publish file list."""
        self.wolk_device.with_file_management(self.file_directory, 1024)
        os.rmdir(self.file_directory)

        self.wolk_device.message_deserializer.is_file_delete_command = (
            MagicMock(return_value=True)
        )
        self.wolk_device.message_deserializer.parse_file_delete_command = (
            MagicMock(return_value="file")
        )
        self.wolk_device.file_management.handle_file_delete = MagicMock()
        self.wolk_device.file_management.get_file_list = MagicMock(
            return_value=[]
        )
        self.wolk_device.message_factory.make_from_file_list = MagicMock(
            return_value=True
        )
        self.wolk_device.connectivity_service.publish = MagicMock(
            return_value=False
        )
        self.wolk_device.message_queue.put = MagicMock()

        self.wolk_device._on_file_management_message(self.message)

        self.wolk_device.message_queue.put.assert_called_once()

    def test_on_file_management_message_file_delete_publishes(self):
        """Test receiving file delete command and send file list."""
        self.wolk_device.with_file_management(self.file_directory, 1024)
        os.rmdir(self.file_directory)

        self.wolk_device.message_deserializer.is_file_delete_command = (
            MagicMock(return_value=True)
        )
        self.wolk_device.message_deserializer.parse_file_delete_command = (
            MagicMock(return_value="file")
        )
        self.wolk_device.file_management.handle_file_delete = MagicMock()
        self.wolk_device.file_management.get_file_list = MagicMock(
            return_value=[]
        )
        self.wolk_device.message_factory.make_from_file_list = MagicMock(
            return_value=True
        )
        self.wolk_device.connectivity_service.publish = MagicMock(
            return_value=True
        )
        self.wolk_device.message_queue.put = MagicMock()

        self.wolk_device._on_file_management_message(self.message)

        self.wolk_device.message_queue.put.assert_not_called()

    def test_on_file_management_message_file_purge_fail_to_publish(self):
        """Test receiving file purge command and fail to publish file list."""
        self.wolk_device.with_file_management(self.file_directory, 1024)
        os.rmdir(self.file_directory)

        self.wolk_device.message_deserializer.is_file_purge_command = (
            MagicMock(return_value=True)
        )
        self.wolk_device.file_management.handle_file_purge = MagicMock()
        self.wolk_device.file_management.get_file_list = MagicMock(
            return_value=[]
        )
        self.wolk_device.message_factory.make_from_file_list = MagicMock(
            return_value=True
        )
        self.wolk_device.connectivity_service.publish = MagicMock(
            return_value=False
        )
        self.wolk_device.message_queue.put = MagicMock()

        self.wolk_device._on_file_management_message(self.message)

        self.wolk_device.message_queue.put.assert_called_once()

    def test_on_file_management_message_file_purge_publishes(self):
        """Test receiving file purge command and send file list."""
        self.wolk_device.with_file_management(self.file_directory, 1024)
        os.rmdir(self.file_directory)

        self.wolk_device.message_deserializer.is_file_purge_command = (
            MagicMock(return_value=True)
        )
        self.wolk_device.file_management.handle_file_purge = MagicMock()
        self.wolk_device.file_management.get_file_list = MagicMock(
            return_value=[]
        )
        self.wolk_device.message_factory.make_from_file_list = MagicMock(
            return_value=True
        )
        self.wolk_device.connectivity_service.publish = MagicMock(
            return_value=True
        )
        self.wolk_device.message_queue.put = MagicMock()

        self.wolk_device._on_file_management_message(self.message)

        self.wolk_device.message_queue.put.assert_not_called()

    def test_on_file_management_message_unkown(self):
        """Test receiving unknown file management message."""
        self.wolk_device.with_file_management(self.file_directory, 1024)
        os.rmdir(self.file_directory)

        self.wolk_device.logger.warning = MagicMock()

        self.wolk_device._on_file_management_message(self.message)

        self.wolk_device.logger.warning.assert_called_once()

    def test_on_firmware_message_no_module_fail_to_publish(self):
        """Test receiving firmware message with no module and fail to publish."""
        self.wolk_device.message_queue.put = MagicMock()
        self.wolk_device.message_factory.make_from_firmware_update_status = (
            MagicMock(return_value=True)
        )

        self.wolk_device.connectivity_service.publish = MagicMock(
            return_value=False
        )
        self.wolk_device._on_firmware_message(self.message)
        self.wolk_device.message_queue.put.assert_called_once()

    def test_on_firmware_message_no_module_fail_publishes(self):
        """Test receiving firmware message with no module and fail to publish."""
        self.wolk_device.logger.warning = MagicMock()

        self.wolk_device.message_queue.put = MagicMock()
        self.wolk_device.message_factory.make_from_firmware_update_status = (
            MagicMock(return_value=True)
        )

        self.wolk_device.connectivity_service.publish = MagicMock(
            return_value=True
        )
        self.wolk_device._on_firmware_message(self.message)
        self.wolk_device.message_queue.put.assert_not_called()

    def test_on_firmware_message_firmware_install_no_path_fail_to_publish(
        self,
    ):
        """Test install command non-present file and fail to publish status."""
        self.wolk_device.with_file_management(self.file_directory, 1024)
        os.rmdir(self.file_directory)
        self.firmware_handler.get_current_version = MagicMock(
            return_value="1.0"
        )
        self.wolk_device.with_firmware_update(self.firmware_handler)
        self.wolk_device.message_deserializer.is_firmware_install = MagicMock(
            return_value=True
        )
        self.wolk_device.message_deserializer.parse_firmware_install = (
            MagicMock(return_value=None)
        )
        self.wolk_device.file_management.get_file_path = MagicMock(
            return_value=None
        )
        self.wolk_device.connectivity_service.publish = MagicMock(
            return_value=False
        )
        self.wolk_device.message_queue.put = MagicMock()

        self.wolk_device._on_firmware_message(self.message)

        self.wolk_device.message_queue.put.assert_called_once()

    def test_on_firmware_message_firmware_install_no_path_publishes(
        self,
    ):
        """Test install command non-present file and publishes status."""
        self.wolk_device.with_file_management(self.file_directory, 1024)
        os.rmdir(self.file_directory)
        self.firmware_handler.get_current_version = MagicMock(
            return_value="1.0"
        )
        self.wolk_device.with_firmware_update(self.firmware_handler)
        self.wolk_device.connectivity_service.publish = MagicMock(
            return_value=True
        )
        self.wolk_device.message_deserializer.is_firmware_install = MagicMock(
            return_value=True
        )
        self.wolk_device.message_deserializer.parse_firmware_install = (
            MagicMock(return_value=None)
        )
        self.wolk_device.file_management.get_file_path = MagicMock(
            return_value=None
        )
        self.wolk_device.message_queue.put = MagicMock()

        self.wolk_device._on_firmware_message(self.message)

        self.wolk_device.message_queue.put.assert_not_called()

    def test_on_firmware_message_firmware_install_with_path_calls_intsall(
        self,
    ):
        """Test install command present file calls handle install."""
        self.wolk_device.with_file_management(self.file_directory, 1024)
        os.rmdir(self.file_directory)
        self.firmware_handler.get_current_version = MagicMock(
            return_value="1.0"
        )
        self.wolk_device.with_firmware_update(self.firmware_handler)
        self.wolk_device.connectivity_service.publish = MagicMock(
            return_value=True
        )
        self.wolk_device.message_deserializer.is_firmware_install = MagicMock(
            return_value=True
        )
        self.wolk_device.message_deserializer.parse_firmware_install = (
            MagicMock(return_value=None)
        )
        self.wolk_device.file_management.get_file_path = MagicMock(
            return_value="file"
        )
        self.wolk_device.firmware_update.handle_install = MagicMock()
        self.wolk_device._on_firmware_message(self.message)

        self.wolk_device.firmware_update.handle_install.assert_called_once_with(
            "file"
        )

    def test_on_firmware_message_firmware_abort(self):
        """Test abort command calls handle abort."""
        self.wolk_device.with_file_management(self.file_directory, 1024)
        os.rmdir(self.file_directory)
        self.firmware_handler.get_current_version = MagicMock(
            return_value="1.0"
        )
        self.wolk_device.with_firmware_update(self.firmware_handler)
        self.wolk_device.message_deserializer.is_firmware_abort = MagicMock(
            return_value=True
        )
        self.wolk_device.firmware_update.handle_abort = MagicMock()
        self.wolk_device._on_firmware_message(self.message)

        self.wolk_device.firmware_update.handle_abort.assert_called_once_with()

    def test_on_firmware_message_firmware_version_request_pulishes(self):
        """Test receiving version request and publishes response."""
        self.wolk_device.with_file_management(self.file_directory, 1024)
        os.rmdir(self.file_directory)
        self.firmware_handler.get_current_version = MagicMock(
            return_value="1.0"
        )
        self.wolk_device.with_firmware_update(self.firmware_handler)
        self.wolk_device.message_deserializer.is_firmware_version_request = (
            MagicMock(return_value=True)
        )
        self.wolk_device.message_factory.make_from_firmware_version_response = MagicMock(
            return_value="1.0"
        )
        self.wolk_device.connectivity_service.publish = MagicMock(
            return_value=True
        )
        self.wolk_device.message_queue.put = MagicMock()

        self.wolk_device._on_firmware_message(self.message)

        self.wolk_device.message_queue.put.assert_not_called()

    def test_on_firmware_message_unknown(self):
        """Test receiving unknown firmware message."""
        self.wolk_device.logger.setLevel(logging.WARNING)
        self.wolk_device.with_file_management(self.file_directory, 1024)
        os.rmdir(self.file_directory)
        self.firmware_handler.get_current_version = MagicMock(
            return_value="1.0"
        )
        self.wolk_device.with_firmware_update(self.firmware_handler)
        self.wolk_device.logger.warning = MagicMock()

        self.wolk_device._on_firmware_message(self.message)

        self.wolk_device.logger.warning.assert_called_once()

    def test_on_package_request_fails_to_publish(self):
        """Test making package request and failing to send it."""
        self.wolk_device.with_file_management(self.file_directory, 1024)
        os.rmdir(self.file_directory)
        self.wolk_device.message_factory.make_from_package_request = MagicMock(
            return_value=True
        )
        self.wolk_device.connectivity_service.publish = MagicMock(
            return_value=False
        )
        self.wolk_device.message_queue.put = MagicMock()

        self.wolk_device._on_package_request("file", 0)

        self.wolk_device.message_queue.put.assert_called_once()

    def test_on_package_request_publishes(self):
        """Test making package request and publishing it."""
        self.wolk_device.with_file_management(self.file_directory, 1024)
        os.rmdir(self.file_directory)
        self.wolk_device.message_factory.make_from_package_request = MagicMock(
            return_value=True
        )
        self.wolk_device.connectivity_service.publish = MagicMock(
            return_value=True
        )
        self.wolk_device.message_queue.put = MagicMock()

        self.wolk_device._on_package_request("file", 0)

        self.wolk_device.message_queue.put.assert_not_called()

    def test_on_firmware_update_status_not_connected(self):
        """Test on firmware update status call when not connected."""
        self.wolk_device.with_file_management(self.file_directory, 1024)
        os.rmdir(self.file_directory)
        self.firmware_handler.get_current_version = MagicMock(
            return_value="1.0"
        )
        self.wolk_device.with_firmware_update(self.firmware_handler)

        status = FirmwareUpdateStatus(FirmwareUpdateStatusType.INSTALLING)
        self.wolk_device.connectivity_service.is_connected = MagicMock(
            return_value=False
        )
        self.wolk_device.connectivity_service.publish = MagicMock()
        self.wolk_device.message_factory.make_from_firmware_update_status = (
            MagicMock(return_value=True)
        )
        self.wolk_device._on_firmware_update_status(status)

        self.wolk_device.connectivity_service.publish.assert_not_called()

    def test_on_firmware_update_status_fail_to_publish(self):
        """Test on firmware update status and fail to publish."""
        self.wolk_device.with_file_management(self.file_directory, 1024)
        os.rmdir(self.file_directory)
        self.firmware_handler.get_current_version = MagicMock(
            return_value="1.0"
        )
        self.wolk_device.with_firmware_update(self.firmware_handler)

        status = FirmwareUpdateStatus(FirmwareUpdateStatusType.INSTALLING)
        self.wolk_device.connectivity_service.is_connected = MagicMock(
            return_value=True
        )
        self.wolk_device.connectivity_service.publish = MagicMock(
            return_value=False
        )
        self.wolk_device.message_factory.make_from_firmware_update_status = (
            MagicMock(return_value=True)
        )
        self.wolk_device.message_queue.put = MagicMock()
        self.wolk_device._on_firmware_update_status(status)

        self.wolk_device.message_queue.put.assert_called_once()

    def test_on_firmware_update_status_publishes(self):
        """Test on firmware update status and publishes the message."""
        self.wolk_device.with_file_management(self.file_directory, 1024)
        os.rmdir(self.file_directory)
        self.firmware_handler.get_current_version = MagicMock(
            return_value="1.0"
        )
        self.wolk_device.with_firmware_update(self.firmware_handler)

        status = FirmwareUpdateStatus(FirmwareUpdateStatusType.INSTALLING)
        self.wolk_device.connectivity_service.is_connected = MagicMock(
            return_value=True
        )
        self.wolk_device.connectivity_service.publish = MagicMock(
            return_value=True
        )
        self.wolk_device.message_factory.make_from_firmware_update_status = (
            MagicMock(return_value=True)
        )
        self.wolk_device.message_queue.put = MagicMock()
        self.wolk_device._on_firmware_update_status(status)

        self.wolk_device.message_queue.put.assert_not_called()

    def test_on_firmware_update_status_completed_not_connected(self):
        """Test on firmware status completed and not connected."""
        self.wolk_device.with_file_management(self.file_directory, 1024)
        os.rmdir(self.file_directory)
        self.firmware_handler.get_current_version = MagicMock(
            return_value="1.0"
        )
        self.wolk_device.with_firmware_update(self.firmware_handler)

        status = FirmwareUpdateStatus(FirmwareUpdateStatusType.SUCCESS)
        self.wolk_device.connectivity_service.is_connected = MagicMock(
            return_value=False
        )
        self.wolk_device.connectivity_service.publish = MagicMock()
        self.wolk_device.message_factory.make_from_firmware_update_status = (
            MagicMock(return_value=True)
        )
        self.wolk_device.message_factory.make_from_firmware_version_update = (
            MagicMock(return_value=True)
        )
        self.wolk_device._on_firmware_update_status(status)

        self.wolk_device.connectivity_service.publish.assert_not_called()

    def test_on_firmware_update_status_completed_fail_to_publish(self):
        """Test on firmware status completed and fail to publish."""
        self.wolk_device.with_file_management(self.file_directory, 1024)
        os.rmdir(self.file_directory)
        self.firmware_handler.get_current_version = MagicMock(
            return_value="1.0"
        )
        self.wolk_device.with_firmware_update(self.firmware_handler)

        status = FirmwareUpdateStatus(FirmwareUpdateStatusType.SUCCESS)
        self.wolk_device.connectivity_service.is_connected = MagicMock(
            return_value=True
        )
        self.wolk_device.connectivity_service.publish = MagicMock(
            return_value=False
        )
        self.wolk_device.message_factory.make_from_firmware_update_status = (
            MagicMock(return_value=True)
        )
        self.wolk_device.message_factory.make_from_firmware_version_update = (
            MagicMock(return_value=True)
        )
        self.wolk_device.message_queue.put = MagicMock()
        self.wolk_device._on_firmware_update_status(status)

        self.assertEqual(2, self.wolk_device.message_queue.put.call_count)

    def test_on_firmware_update_status_completed_publishes(self):
        """Test on firmware status completed and publishes message."""
        self.wolk_device.with_file_management(self.file_directory, 1024)
        os.rmdir(self.file_directory)
        self.firmware_handler.get_current_version = MagicMock(
            return_value="1.0"
        )
        self.wolk_device.with_firmware_update(self.firmware_handler)

        status = FirmwareUpdateStatus(FirmwareUpdateStatusType.SUCCESS)
        self.wolk_device.connectivity_service.is_connected = MagicMock(
            return_value=True
        )
        self.wolk_device.connectivity_service.publish = MagicMock(
            return_value=True
        )
        self.wolk_device.message_factory.make_from_firmware_update_status = (
            MagicMock(return_value=True)
        )
        self.wolk_device.message_factory.make_from_firmware_version_update = (
            MagicMock(return_value=True)
        )
        self.wolk_device.message_queue.put = MagicMock()
        self.wolk_device._on_firmware_update_status(status)
        self.wolk_device.message_queue.put.assert_not_called()

    def test_on_file_upload_status_fail_to_publish(self):
        """Test on file upload status and fail to publish message."""
        self.wolk_device.with_file_management(self.file_directory, 1024)
        os.rmdir(self.file_directory)

        self.wolk_device.message_factory.make_from_file_management_status = (
            MagicMock(return_value=True)
        )
        self.wolk_device.connectivity_service.publish = MagicMock(
            return_value=False
        )
        status = FileManagementStatus(FileManagementStatusType.FILE_TRANSFER)
        self.wolk_device.message_queue.put = MagicMock()

        self.wolk_device._on_file_upload_status(self.file_name, status)

        self.wolk_device.message_queue.put.assert_called_once()

    def test_on_file_upload_status_publishes(self):
        """Test on file upload status and publishes message."""
        self.wolk_device.with_file_management(self.file_directory, 1024)
        os.rmdir(self.file_directory)

        self.wolk_device.message_factory.make_from_file_management_status = (
            MagicMock(return_value=True)
        )
        self.wolk_device.connectivity_service.publish = MagicMock(
            return_value=True
        )
        status = FileManagementStatus(FileManagementStatusType.FILE_TRANSFER)
        self.wolk_device.message_queue.put = MagicMock()

        self.wolk_device._on_file_upload_status(self.file_name, status)

        self.wolk_device.message_queue.put.assert_not_called()

    def test_on_file_upload_status_file_ready_fail_to_publish(self):
        """Test on file upload status and fail to publish message."""
        self.wolk_device.with_file_management(self.file_directory, 1024)
        os.rmdir(self.file_directory)

        self.wolk_device.message_factory.make_from_file_management_status = (
            MagicMock(return_value=True)
        )
        self.wolk_device.connectivity_service.publish = MagicMock(
            return_value=False
        )
        self.wolk_device.message_factory.make_from_file_list = MagicMock(
            return_value=True
        )
        self.wolk_device.file_management.get_file_list = MagicMock(
            return_value=[]
        )
        status = FileManagementStatus(FileManagementStatusType.FILE_READY)
        self.wolk_device.message_queue.put = MagicMock()

        self.wolk_device._on_file_upload_status(self.file_name, status)

        self.assertEqual(2, self.wolk_device.message_queue.put.call_count)

    def test_on_file_upload_status_file_ready_published(self):
        """Test on file upload status and publishes message."""
        self.wolk_device.with_file_management(self.file_directory, 1024)
        os.rmdir(self.file_directory)

        self.wolk_device.message_factory.make_from_file_management_status = (
            MagicMock(return_value=True)
        )
        self.wolk_device.connectivity_service.publish = MagicMock(
            return_value=True
        )
        self.wolk_device.message_factory.make_from_file_list = MagicMock(
            return_value=True
        )
        self.wolk_device.file_management.get_file_list = MagicMock(
            return_value=[]
        )
        status = FileManagementStatus(FileManagementStatusType.FILE_READY)
        self.wolk_device.message_queue.put = MagicMock()

        self.wolk_device._on_file_upload_status(self.file_name, status)

        self.wolk_device.message_queue.put.assert_not_called()

    def test_on_file_url_status_fail_to_publish(self):
        """Test on file URL status and fail to publish update."""
        self.wolk_device.with_file_management(self.file_directory, 1024)
        os.rmdir(self.file_directory)

        self.wolk_device.message_factory.make_from_file_url_status = MagicMock(
            return_value=True
        )
        self.wolk_device.connectivity_service.publish = MagicMock(
            return_value=False
        )
        status = FileManagementStatus(FileManagementStatusType.FILE_TRANSFER)
        self.wolk_device.message_queue.put = MagicMock()

        self.wolk_device._on_file_url_status(self.file_url, status)

        self.wolk_device.message_queue.put.assert_called_once()

    def test_on_file_url_status_publishes(self):
        """Test on file URL status and publishes update."""
        self.wolk_device.with_file_management(self.file_directory, 1024)
        os.rmdir(self.file_directory)

        self.wolk_device.message_factory.make_from_file_url_status = MagicMock(
            return_value=True
        )
        self.wolk_device.connectivity_service.publish = MagicMock(
            return_value=True
        )
        status = FileManagementStatus(FileManagementStatusType.FILE_TRANSFER)
        self.wolk_device.message_queue.put = MagicMock()

        self.wolk_device._on_file_url_status(self.file_url, status)

        self.wolk_device.message_queue.put.assert_not_called()

    def test_on_file_url_status_with_file_name_fail_to_publish(self):
        """Test on URL upload status and fail to publish message."""
        self.wolk_device.with_file_management(self.file_directory, 1024)
        os.rmdir(self.file_directory)

        self.wolk_device.message_factory.make_from_file_management_status = (
            MagicMock(return_value=True)
        )
        self.wolk_device.connectivity_service.publish = MagicMock(
            return_value=False
        )
        self.wolk_device.message_factory.make_from_file_url_status = MagicMock(
            return_value=True
        )
        self.wolk_device.file_management.get_file_list = MagicMock(
            return_value=[]
        )
        status = FileManagementStatus(FileManagementStatusType.FILE_READY)
        self.wolk_device.message_queue.put = MagicMock()

        self.wolk_device._on_file_url_status(
            self.file_url, status, self.file_name
        )

        self.assertEqual(2, self.wolk_device.message_queue.put.call_count)

    def test_on_file_url_status_with_file_name_publishes(self):
        """Test on URL upload status and publishes message."""
        self.wolk_device.with_file_management(self.file_directory, 1024)
        os.rmdir(self.file_directory)

        self.wolk_device.message_factory.make_from_file_management_status = (
            MagicMock(return_value=True)
        )
        self.wolk_device.connectivity_service.publish = MagicMock(
            return_value=True
        )
        self.wolk_device.message_factory.make_from_file_url_status = MagicMock(
            return_value=True
        )
        self.wolk_device.file_management.get_file_list = MagicMock(
            return_value=[]
        )
        status = FileManagementStatus(FileManagementStatusType.FILE_READY)
        self.wolk_device.message_queue.put = MagicMock()

        self.wolk_device._on_file_url_status(
            self.file_url, status, self.file_name
        )

        self.wolk_device.message_queue.put.assert_not_called()

    def test_on_parameters_message(self):
        """Test on parameters message received."""
        self.wolk_device.logger.setLevel(logging.WARNING)
        self.wolk_device.message_deserializer.parse_parameters = MagicMock(
            return_value=[{}]
        )
        self.wolk_device.message_deserializer.is_parameters = MagicMock(
            return_value=True
        )
        self.wolk_device.parameters = MagicMock()
        self.wolk_device.parameters.update = MagicMock()
        self.wolk_device._on_inbound_message(Message("test"))

        self.wolk_device.parameters.update.assert_called_once()

    def test_on_feed_values_message_no_handler(self):
        """Test on feed values message received with no handler."""
        self.wolk_device.logger.setLevel(logging.WARNING)
        self.wolk_device.logger.warning = MagicMock()
        self.wolk_device.message_deserializer.is_parameters = MagicMock(
            return_value=False
        )
        self.wolk_device.message_deserializer.is_feed_values = MagicMock(
            return_value=True
        )
        self.wolk_device.incoming_feed_value_handler = None
        self.wolk_device._on_inbound_message(Message("test"))

        self.wolk_device.logger.warning.assert_called_once()

    def test_on_feed_values_message_fail_to_parse(self):
        """Test on feed values message that failed to parse."""
        self.wolk_device.logger.setLevel(logging.WARNING)
        self.wolk_device.logger.warning = MagicMock()
        self.wolk_device.message_deserializer.is_parameters = MagicMock(
            return_value=False
        )
        self.wolk_device.message_deserializer.is_feed_values = MagicMock(
            return_value=True
        )
        self.wolk_device.message_deserializer.parse_feed_values = MagicMock(
            return_value=None
        )
        self.wolk_device.incoming_feed_value_handler = True
        self.wolk_device._on_inbound_message(Message("test"))

        self.wolk_device.logger.warning.assert_called_once()

    def test_on_feed_values_message(self):
        """Test on feed values message."""
        self.wolk_device.logger.setLevel(logging.WARNING)
        self.wolk_device.logger.warning = MagicMock()
        self.wolk_device.message_deserializer.is_parameters = MagicMock(
            return_value=False
        )
        self.wolk_device.message_deserializer.is_feed_values = MagicMock(
            return_value=True
        )
        self.wolk_device.message_deserializer.parse_feed_values = MagicMock(
            return_value=True
        )
        self.wolk_device.incoming_feed_value_handler = MagicMock()
        self.wolk_device._on_inbound_message(Message("test"))

        self.wolk_device.incoming_feed_value_handler.assert_called_once_with(
            True
        )
        self.wolk_device.logger.warning.assert_not_called()

    def test_request_timestamp_when_none(self):
        """Test request timestamp returns none."""
        self.assertIsNone(self.wolk_device.request_timestamp())

    def test_request_timestamp_when_not_none(self):
        """Test request timestamp returns value."""
        self.wolk_device.last_platform_timestamp = 1

        self.assertIsNotNone(self.wolk_device.request_timestamp())

    def test_add_feed_value(self):
        """Test add feed value."""
        self.wolk_device.logger.setLevel(logging.CRITICAL)
        self.wolk_device.message_queue.put = MagicMock()
        self.wolk_device.message_factory.make_from_feed_value = MagicMock(
            return_value=True
        )

        self.wolk_device.add_feed_value("foo", "bar")

        self.wolk_device.message_queue.put.assert_called_once()

    def test_pull_parameters_not_pull_device(self):
        """Test pull parameters for a device that isn't PULL."""
        self.wolk_device.logger.setLevel(logging.WARNING)
        self.wolk_device.logger.warning = MagicMock()

        self.wolk_device.pull_parameters()

        self.wolk_device.logger.warning.assert_called_once()

    def test_pull_parameters_not_connected(self):
        """Test pull parameters when not connected."""
        self.wolk_device.logger.setLevel(logging.WARNING)
        self.wolk_device.logger.warning = MagicMock()
        self.wolk_device.connectivity_service.is_connected = MagicMock(
            return_value=False
        )
        self.wolk_device.device.data_delivery = DataDelivery.PULL

        self.wolk_device.pull_parameters()

        self.wolk_device.logger.warning.assert_called_once()

    def test_pull_parameters_fails_to_publish(self):
        """Test pull parameters fails to publish."""
        self.wolk_device.logger.setLevel(logging.WARNING)
        self.wolk_device.logger.warning = MagicMock()
        self.wolk_device.connectivity_service.is_connected = MagicMock(
            return_value=True
        )
        self.wolk_device.device.data_delivery = DataDelivery.PULL
        self.wolk_device.message_factory.make_pull_parameters = MagicMock(
            return_value=True
        )
        self.wolk_device.connectivity_service.publish = MagicMock(
            return_value=False
        )

        self.wolk_device.pull_parameters()

        self.wolk_device.logger.warning.assert_called_once()

    def test_pull_parameters_publishes(self):
        """Test pull parameters publishes."""
        self.wolk_device.logger.setLevel(logging.WARNING)
        self.wolk_device.logger.warning = MagicMock()
        self.wolk_device.connectivity_service.is_connected = MagicMock(
            return_value=True
        )
        self.wolk_device.device.data_delivery = DataDelivery.PULL
        self.wolk_device.message_factory.make_pull_parameters = MagicMock(
            return_value=True
        )
        self.wolk_device.connectivity_service.publish = MagicMock(
            return_value=True
        )

        self.wolk_device.pull_parameters()

        self.wolk_device.logger.warning.assert_not_called()

    def test_pull_feed_values_not_pull_device(self):
        """Test pull feed values for a device that isn't PULL."""
        self.wolk_device.logger.setLevel(logging.WARNING)
        self.wolk_device.logger.warning = MagicMock()

        self.wolk_device.pull_feed_values()

        self.wolk_device.logger.warning.assert_called_once()

    def test_pull_feed_values_not_connected(self):
        """Test pull feed values when not connected."""
        self.wolk_device.logger.setLevel(logging.WARNING)
        self.wolk_device.logger.warning = MagicMock()
        self.wolk_device.connectivity_service.is_connected = MagicMock(
            return_value=False
        )
        self.wolk_device.device.data_delivery = DataDelivery.PULL

        self.wolk_device.pull_feed_values()

        self.wolk_device.logger.warning.assert_called_once()

    def test_pull_feed_values_fails_to_publish(self):
        """Test pull feed values fails to publish."""
        self.wolk_device.logger.setLevel(logging.WARNING)
        self.wolk_device.logger.warning = MagicMock()
        self.wolk_device.connectivity_service.is_connected = MagicMock(
            return_value=True
        )
        self.wolk_device.device.data_delivery = DataDelivery.PULL
        self.wolk_device.message_factory.make_pull_feed_values = MagicMock(
            return_value=True
        )
        self.wolk_device.connectivity_service.publish = MagicMock(
            return_value=False
        )

        self.wolk_device.pull_feed_values()

        self.wolk_device.logger.warning.assert_called_once()

    def test_pull_feed_values_publishes(self):
        """Test pull feed values publishes."""
        self.wolk_device.logger.setLevel(logging.WARNING)
        self.wolk_device.logger.warning = MagicMock()
        self.wolk_device.connectivity_service.is_connected = MagicMock(
            return_value=True
        )
        self.wolk_device.device.data_delivery = DataDelivery.PULL
        self.wolk_device.message_factory.make_pull_feed_values = MagicMock(
            return_value=True
        )
        self.wolk_device.connectivity_service.publish = MagicMock(
            return_value=True
        )

        self.wolk_device.pull_feed_values()

        self.wolk_device.logger.warning.assert_not_called()

    def test_register_feed_not_connected(self):
        """Test registering a feed when not connected."""
        self.wolk_device.logger.setLevel(logging.WARNING)
        self.wolk_device.logger.warning = MagicMock()
        self.wolk_device.connectivity_service.is_connected = MagicMock(
            return_value=False
        )
        self.wolk_device.message_factory.make_feed_registration = MagicMock(
            return_value=True
        )

        self.wolk_device.register_feed("foo", "bar", FeedType.IN, Unit.CELSIUS)

        self.wolk_device.logger.warning.assert_called_once()

    def test_register_feed_fails_to_publish(self):
        """Test registering a feed when not connected."""
        self.wolk_device.logger.setLevel(logging.WARNING)
        self.wolk_device.logger.warning = MagicMock()
        self.wolk_device.message_queue.put = MagicMock()
        self.wolk_device.connectivity_service.is_connected = MagicMock(
            return_value=True
        )
        self.wolk_device.message_factory.make_feed_registration = MagicMock(
            return_value=True
        )
        self.wolk_device.connectivity_service.publish = MagicMock(
            return_value=False
        )

        self.wolk_device.register_feed("foo", "bar", FeedType.IN, Unit.CELSIUS)

        self.wolk_device.logger.warning.assert_called_once()
        self.wolk_device.message_queue.put.assert_called_once()

    def test_register_feed_custom_unit(self):
        """Test registering a feed with custom unit."""
        self.wolk_device.logger.setLevel(logging.WARNING)
        self.wolk_device.logger.warning = MagicMock()
        self.wolk_device.message_queue.put = MagicMock()
        self.wolk_device.connectivity_service.is_connected = MagicMock(
            return_value=True
        )
        self.wolk_device.message_factory.make_feed_registration = MagicMock(
            return_value=True
        )
        self.wolk_device.connectivity_service.publish = MagicMock(
            return_value=True
        )

        self.wolk_device.register_feed("foo", "bar", FeedType.IN, "custom")

        self.wolk_device.logger.warning.assert_called_once()

    def test_remove_feed_not_connected(self):
        """Test removing a feed when not connected."""
        self.wolk_device.logger.setLevel(logging.WARNING)
        self.wolk_device.logger.warning = MagicMock()
        self.wolk_device.message_queue.put = MagicMock()
        self.wolk_device.connectivity_service.is_connected = MagicMock(
            return_value=False
        )
        self.wolk_device.message_factory.make_feed_removal = MagicMock(
            return_value=True
        )

        self.wolk_device.remove_feed("foo")

        self.wolk_device.logger.warning.assert_called_once()
        self.wolk_device.message_queue.put.assert_called_once()

    def test_remove_feed_fail_to_publish(self):
        """Test removing a feed fails to publish."""
        self.wolk_device.logger.setLevel(logging.WARNING)
        self.wolk_device.logger.warning = MagicMock()
        self.wolk_device.message_queue.put = MagicMock()
        self.wolk_device.connectivity_service.is_connected = MagicMock(
            return_value=True
        )
        self.wolk_device.message_factory.make_feed_removal = MagicMock(
            return_value=True
        )
        self.wolk_device.connectivity_service.publish = MagicMock(
            return_value=False
        )

        self.wolk_device.remove_feed("foo")

        self.wolk_device.logger.warning.assert_called_once()
        self.wolk_device.message_queue.put.assert_called_once()

    def test_remove_feed_publishes(self):
        """Test remove feed request publishes."""
        self.wolk_device.logger.setLevel(logging.WARNING)
        self.wolk_device.logger.warning = MagicMock()
        self.wolk_device.message_queue.put = MagicMock()
        self.wolk_device.connectivity_service.is_connected = MagicMock(
            return_value=True
        )
        self.wolk_device.message_factory.make_feed_removal = MagicMock(
            return_value=True
        )
        self.wolk_device.connectivity_service.publish = MagicMock(
            return_value=True
        )

        self.wolk_device.remove_feed("foo")

        self.wolk_device.logger.warning.assert_not_called()
        self.wolk_device.message_queue.put.assert_not_called()

    def test_register_attribute_not_connected(self):
        """Test registering attribute when not connected."""
        self.wolk_device.logger.setLevel(logging.WARNING)
        self.wolk_device.logger.warning = MagicMock()
        self.wolk_device.message_queue.put = MagicMock()
        self.wolk_device.connectivity_service.is_connected = MagicMock(
            return_value=False
        )
        self.wolk_device.message_factory.make_attribute_registration = (
            MagicMock(return_value=True)
        )

        self.wolk_device.register_attribute("foo", DataType.STRING, "bar")

        self.wolk_device.logger.warning.assert_called_once()
        self.wolk_device.message_queue.put.assert_called_once()

    def test_register_attribute_fails_to_publish(self):
        """Test registering attribute that fails to publish."""
        self.wolk_device.logger.setLevel(logging.WARNING)
        self.wolk_device.logger.warning = MagicMock()
        self.wolk_device.message_queue.put = MagicMock()
        self.wolk_device.connectivity_service.is_connected = MagicMock(
            return_value=True
        )
        self.wolk_device.connectivity_service.publish = MagicMock(
            return_value=False
        )
        self.wolk_device.message_factory.make_attribute_registration = (
            MagicMock(return_value=True)
        )

        self.wolk_device.register_attribute("foo", DataType.STRING, "bar")

        self.wolk_device.logger.warning.assert_called_once()
        self.wolk_device.message_queue.put.assert_called_once()

    def test_register_attribute_publishes(self):
        """Test registering attribute publishes."""
        self.wolk_device.logger.setLevel(logging.WARNING)
        self.wolk_device.logger.warning = MagicMock()
        self.wolk_device.message_queue.put = MagicMock()
        self.wolk_device.connectivity_service.is_connected = MagicMock(
            return_value=True
        )
        self.wolk_device.connectivity_service.publish = MagicMock(
            return_value=True
        )
        self.wolk_device.message_factory.make_attribute_registration = (
            MagicMock(return_value=True)
        )

        self.wolk_device.register_attribute("foo", DataType.STRING, "bar")

        self.wolk_device.logger.warning.assert_not_called()
        self.wolk_device.message_queue.put.assert_not_called()



"""Tests for OSFirmwareUpdate."""
#   Copyright 2020 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import logging
import os
import sys
import unittest
from unittest.mock import MagicMock

sys.path.append("..")  # noqa

from wolk.os_firmware_update import OSFirmwareUpdate
from wolk.interface.firmware_handler import FirmwareHandler
from wolk.model.firmware_update_error_type import FirmwareUpdateErrorType
from wolk.model.firmware_update_status import FirmwareUpdateStatus
from wolk.model.firmware_update_status_type import FirmwareUpdateStatusType


class TestOSFirmwareUpdate(unittest.TestCase):
    """Tests for OSFirmwareUpdate class."""

    class MockFirmwareHandler(FirmwareHandler):
        """Mock firmware installer class that whose methods will be mocked."""

        def install_firmware(self, firmware_file_path: str) -> None:
            """
            Handle the installation of the firmware file.

            :param firmware_file_path: Path where the firmware file is located
            :type firmware_file_path: str
            """
            raise NotImplementedError()

        def get_current_version(self) -> str:
            """
            Obtain device's current firmware version.

            :returns: version
            :rtype: str
            """
            raise NotImplementedError()

    def test_invalid_firmware_handler(self):
        """Test passing an invalid firmware handler."""
        mock_status_callback = MagicMock()

        self.assertRaises(
            ValueError, OSFirmwareUpdate, 1, mock_status_callback
        )

    def test_get_firmware_version(self):
        """Test getting firmware version from firmware handler."""
        mock_status_callback = MagicMock()
        mock_firmware_handler = self.MockFirmwareHandler()
        firmware_version = "1.0"
        mock_firmware_handler.get_current_version = MagicMock(
            return_value=firmware_version
        )

        firmware_update = OSFirmwareUpdate(
            mock_firmware_handler, mock_status_callback
        )

        self.assertEqual(
            firmware_version, firmware_update.get_current_version()
        )

    def test_handle_install_not_idle(self):
        """Test receiving install command when module not idle."""
        mock_status_callback = MagicMock()
        mock_firmware_handler = self.MockFirmwareHandler()

        firmware_update = OSFirmwareUpdate(
            mock_firmware_handler, mock_status_callback
        )
        firmware_update.logger.setLevel(logging.CRITICAL)
        firmware_update.logger.warning = MagicMock()

        firmware_update.current_status = 1  # Not None

        firmware_update.handle_install("some_file")

        firmware_update.logger.warning.assert_called_once()

    def test_handle_install_existing_version_file(self):
        """Test receiving install command when version file exists."""
        mock_status_callback = MagicMock()
        mock_firmware_handler = self.MockFirmwareHandler()

        firmware_update = OSFirmwareUpdate(
            mock_firmware_handler, mock_status_callback
        )
        firmware_update.logger.setLevel(logging.CRITICAL)

        file_handle = open("last_firmware_version.txt", "w")
        expected_status = FirmwareUpdateStatus(
            FirmwareUpdateStatusType.ERROR,
            FirmwareUpdateErrorType.UNKNOWN,
        )

        firmware_update.handle_install("some_file")

        firmware_update.status_callback.assert_called_once_with(
            expected_status
        )

        file_handle.close()
        os.remove("last_firmware_version.txt")

    def test_handle_install_file_not_present(self):
        """Test receiving install command when file does not exist."""
        mock_status_callback = MagicMock()
        mock_firmware_handler = self.MockFirmwareHandler()
        firmware_version = "1.0"
        mock_firmware_handler.get_current_version = MagicMock(
            return_value=firmware_version
        )

        firmware_update = OSFirmwareUpdate(
            mock_firmware_handler, mock_status_callback
        )
        firmware_update.logger.setLevel(logging.CRITICAL)

        expected_status = FirmwareUpdateStatus(
            FirmwareUpdateStatusType.ERROR,
            FirmwareUpdateErrorType.UNKNOWN_FILE,
        )

        firmware_update.handle_install("some_file")

        firmware_update.status_callback.assert_called_once_with(
            expected_status
        )

    def test_handle_install_file_present(self):
        """Test receiving install command and file exists on device."""
        mock_status_callback = MagicMock()
        mock_firmware_handler = self.MockFirmwareHandler()
        mock_firmware_handler.get_current_version = MagicMock(
            return_value="1.0"
        )
        mock_firmware_handler.install_firmware = MagicMock()

        firmware_update = OSFirmwareUpdate(
            mock_firmware_handler, mock_status_callback
        )
        firmware_update.logger.setLevel(logging.CRITICAL)
        expected_status = FirmwareUpdateStatus(
            FirmwareUpdateStatusType.INSTALLING
        )

        file_handle = open("test_file", "w")
        firmware_update.handle_install("test_file")

        firmware_update.status_callback.assert_called_once_with(
            expected_status
        )
        file_handle.close()
        os.remove("test_file")
        os.remove("last_firmware_version.txt")

    def test_handle_abort_when_not_idle(self):
        """Test receiving the abort command when module not idle."""
        mock_status_callback = MagicMock()
        mock_firmware_handler = self.MockFirmwareHandler()
        firmware_update = OSFirmwareUpdate(
            mock_firmware_handler, mock_status_callback
        )
        firmware_update.logger.setLevel(logging.CRITICAL)
        expected_status = FirmwareUpdateStatus(
            FirmwareUpdateStatusType.ABORTED
        )
        firmware_update.current_status = FirmwareUpdateStatus(
            FirmwareUpdateStatusType.INSTALLING
        )
        firmware_update.handle_abort()

        firmware_update.status_callback.assert_called_once_with(
            expected_status
        )

    def test_handle_abort_when_not_idle_and_version_file(self):
        """Test the abort command when not idle and version file exists."""
        mock_status_callback = MagicMock()
        mock_firmware_handler = self.MockFirmwareHandler()
        firmware_update = OSFirmwareUpdate(
            mock_firmware_handler, mock_status_callback
        )
        firmware_update.logger.setLevel(logging.CRITICAL)
        firmware_update.current_status = FirmwareUpdateStatus(
            FirmwareUpdateStatusType.INSTALLING
        )
        file_handle = open("last_firmware_version.txt", "w")
        file_handle.close()
        firmware_update.handle_abort()

        self.assertFalse(os.path.exists("last_firmware_version.txt"))

    def test_report_result_no_stored_file(self):
        """Test reporting result with no stored firmware version."""
        mock_status_callback = MagicMock()
        mock_firmware_handler = self.MockFirmwareHandler()
        firmware_update = OSFirmwareUpdate(
            mock_firmware_handler, mock_status_callback
        )
        firmware_update.logger.setLevel(logging.CRITICAL)

        firmware_update.logger.debug = MagicMock()

        firmware_update.report_result()

        self.assertEqual(2, firmware_update.logger.debug.call_count)

    def test_report_result_unchanged_version(self):
        """Test reporting result with unchanged version."""
        mock_status_callback = MagicMock()
        mock_firmware_handler = self.MockFirmwareHandler()
        firmware_version = "1.0"
        mock_firmware_handler.get_current_version = MagicMock(
            return_value=firmware_version
        )
        firmware_update = OSFirmwareUpdate(
            mock_firmware_handler, mock_status_callback
        )
        firmware_update.logger.setLevel(logging.CRITICAL)
        with open("last_firmware_version.txt", "w") as file:
            file.write(firmware_update.firmware_handler.get_current_version())

        expected_status = FirmwareUpdateStatus(
            FirmwareUpdateStatusType.ERROR,
            FirmwareUpdateErrorType.INSTALLATION_FAILED,
        )

        firmware_update.report_result()

        firmware_update.status_callback.assert_called_once_with(
            expected_status
        )

    def test_report_result_changed_version(self):
        """Test reporting result with changed version."""
        mock_status_callback = MagicMock()
        mock_firmware_handler = self.MockFirmwareHandler()
        firmware_version = "1.0"
        mock_firmware_handler.get_current_version = MagicMock(
            return_value=firmware_version
        )
        firmware_update = OSFirmwareUpdate(
            mock_firmware_handler, mock_status_callback
        )
        firmware_update.logger.setLevel(logging.CRITICAL)
        with open("last_firmware_version.txt", "w") as file:
            file.write(firmware_update.firmware_handler.get_current_version())

        firmware_version = "2.0"
        mock_firmware_handler.get_current_version = MagicMock(
            return_value=firmware_version
        )
        expected_status = FirmwareUpdateStatus(
            FirmwareUpdateStatusType.SUCCESS
        )

        firmware_update.report_result()

        firmware_update.status_callback.assert_called_once_with(
            expected_status
        )



"""Enumeration of available data types on the Platform."""
#   Copyright 2021 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from enum import Enum


class DataType(Enum):
    """Enumeration of data types on the Platform."""

    STRING = "STRING"
    NUMERIC = "NUMERIC"
    BOOLEAN = "BOOLEAN"
    HEXADECIMAL = "HEXADECIMAL"
    LOCATION = "LOCATION"
    ENUM = "ENUM"



"""
Module that provides connection to WolkAbout IoT Platform.

To start publishing data to the platform
create an instance of Device class with credentials obtained from the platform
and pass it to an instance of WolkConnect class.

For more information about module features visit:
https://github.com/Wolkabout/WolkConnect-Python/tree/master/examples/full_feature_set
"""
__version__ = "5.0.1"
from .interface.connectivity_service import ConnectivityService
from .interface.firmware_handler import FirmwareHandler
from .interface.message_deserializer import MessageDeserializer
from .interface.message_factory import MessageFactory
from .interface.message_queue import MessageQueue
from .logger_factory import logging_config
from .model.data_delivery import DataDelivery
from .model.data_type import DataType
from .model.feed_type import FeedType
from .model.device import Device
from .model.unit import Unit
from .model.file_management_error_type import FileManagementErrorType
from .model.file_management_status import FileManagementStatus
from .model.file_management_status_type import FileManagementStatusType
from .model.file_transfer_package import FileTransferPackage
from .model.firmware_update_error_type import FirmwareUpdateErrorType
from .model.firmware_update_status import FirmwareUpdateStatus
from .model.firmware_update_status_type import FirmwareUpdateStatusType
from .model.message import Message
from .os_file_management import OSFileManagement
from .os_firmware_update import OSFirmwareUpdate
from .wolk_connect import WolkConnect
from .wolkabout_protocol_message_factory import WolkAboutProtocolMessageFactory
from .wolkabout_protocol_message_deserializer import (
    WolkAboutProtocolMessageDeserializer,
)
from .mqtt_connectivity_service import MQTTConnectivityService
from .message_deque import MessageDeque


__all__ = [
    "DataDelivery",
    "DataType",
    "Device",
    "FeedType",
    "FileTransferPackage",
    "FirmwareUpdateStatus",
    "FirmwareUpdateStatusType",
    "FirmwareUpdateErrorType",
    "FileManagementStatus",
    "FileManagementStatusType",
    "FileManagementErrorType",
    "Message",
    "ConnectivityService",
    "FirmwareHandler",
    "MessageDeserializer",
    "MessageFactory",
    "MessageQueue",
    "OSFileManagement",
    "OSFirmwareUpdate",
    "logging_config",
    "Unit",
    "WolkConnect",
    "WolkAboutProtocolMessageFactory",
    "WolkAboutProtocolMessageDeserializer",
    "MQTTConnectivityService",
    "MessageDeque",
    "__version__",
]



"""Tests for MQTTConnectivityService."""
#   Copyright 2020 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import logging
import os
import sys
import unittest
from unittest.mock import MagicMock

from paho.mqtt import client as mqtt

sys.path.append("..")  # noqa

from wolk.model.device import Device
from wolk.model.message import Message
from wolk.mqtt_connectivity_service import MQTTConnectivityService


class MQTTConnectivityServiceTests(unittest.TestCase):
    """Tests for MQTT Connectivity Service."""

    def setUp(self):
        """Set up commonly used test objects."""
        self.device_key = "some_key"
        self.device_password = "some_password"
        self.actuator_references = []
        self.device = Device(
            self.device_key, self.device_password, self.actuator_references
        )
        self.last_will_message = Message("last_will")
        self.topics = []
        self.mqtt_cs = MQTTConnectivityService(
            self.device, self.topics, self.last_will_message
        )
        self.mqtt_cs.logger.setLevel(logging.CRITICAL)

        self.path_to_wolk = (
            os.path.dirname(__file__) + os.sep + ".." + os.sep + "wolk"
        )
        self.ca_crt_path = os.path.join(self.path_to_wolk, "ca.crt")

    def test_is_connected(self):
        """Test is connected method."""
        self.assertFalse(self.mqtt_cs.is_connected())

    def test_set_inbound_message_listener(self):
        """Test setting inbound message listener."""

        def listener():
            raise NotImplementedError

        self.mqtt_cs.set_inbound_message_listener(listener)

        self.assertEqual(listener, self.mqtt_cs.inbound_message_listener)

    def test_on_mqtt_message_no_message(self):
        """Test on mqtt message with no message."""
        self.mqtt_cs.inbound_message_listener = MagicMock()

        self.mqtt_cs._on_mqtt_message(None, None, None)

        self.mqtt_cs.inbound_message_listener.assert_not_called()

    def test_on_mqtt_message_binary_message(self):
        """Test on mqtt message with file chunk message."""
        self.mqtt_cs.inbound_message_listener = MagicMock()
        message = Message("binary", b"")

        self.mqtt_cs._on_mqtt_message(None, None, message)

        self.mqtt_cs.inbound_message_listener.assert_called_once_with(message)

    def test_on_mqtt_message_ordinary_message(self):
        """Test on mqtt message with ordinary message."""
        self.mqtt_cs.inbound_message_listener = MagicMock()
        message = Message("some_topic", "payload")

        self.mqtt_cs._on_mqtt_message(None, None, message)

        self.mqtt_cs.inbound_message_listener.assert_called_once_with(message)

    def test_on_mqtt_connect_rc_0_without_topics(self):
        """Test on mqtt connect with return code 0 without topics."""
        self.mqtt_cs.client.subscribe = MagicMock()
        self.mqtt_cs._on_mqtt_connect(None, None, None, 0)

        self.mqtt_cs.client.subscribe.assert_not_called()

    def test_on_mqtt_connect_rc_0_with_topics(self):
        """Test on mqtt connect with return code 0 with topics."""
        self.mqtt_cs.topics = [1, 2, 3]
        self.mqtt_cs.client = MagicMock()
        self.mqtt_cs.client.subscribe = MagicMock()

        self.mqtt_cs._on_mqtt_connect(None, None, None, 0)

        self.assertEqual(3, self.mqtt_cs.client.subscribe.call_count)

    def test_on_mqtt_connect_rc_1(self):
        """Test on mqtt connect with return code 1."""
        self.mqtt_cs._on_mqtt_connect(None, None, None, 1)

        self.assertEqual(1, self.mqtt_cs.connected_rc)

    def test_on_mqtt_connect_rc_2(self):
        """Test on mqtt connect with return code 2."""
        self.mqtt_cs._on_mqtt_connect(None, None, None, 2)

        self.assertEqual(2, self.mqtt_cs.connected_rc)

    def test_on_mqtt_connect_rc_3(self):
        """Test on mqtt connect with return code 3."""
        self.mqtt_cs._on_mqtt_connect(None, None, None, 3)

        self.assertEqual(3, self.mqtt_cs.connected_rc)

    def test_on_mqtt_connect_rc_4(self):
        """Test on mqtt connect with return code 4."""
        self.mqtt_cs._on_mqtt_connect(None, None, None, 4)

        self.assertEqual(4, self.mqtt_cs.connected_rc)

    def test_on_mqtt_connect_rc_5(self):
        """Test on mqtt connect with return code 5."""
        self.mqtt_cs._on_mqtt_connect(None, None, None, 5)

        self.assertEqual(5, self.mqtt_cs.connected_rc)

    def test_on_mqtt_connect_rc_6(self):
        """Test on mqtt connect with invalid return code 6."""
        self.mqtt_cs._on_mqtt_connect(None, None, None, 6)

        self.assertEqual(None, self.mqtt_cs.connected_rc)

    def test_on_mqtt_disconnect_expected(self):
        """Test on mqtt disconnect with return code 0."""
        self.mqtt_cs.connect = MagicMock()

        self.mqtt_cs._on_mqtt_disconnect(None, None, 0)

        self.mqtt_cs.connect.assert_not_called()

    def test_on_mqtt_disconnect_unexpected(self):
        """Test on mqtt disconnect with return code not 0."""
        self.mqtt_cs.client.reconnect = MagicMock()

        self.mqtt_cs._on_mqtt_disconnect(None, None, 1)

        self.mqtt_cs.client.reconnect.assert_called_once()

    def test_connect_already_connected(self):
        """Test calling connect when already connected."""
        self.mqtt_cs.connected = True

        self.assertTrue(self.mqtt_cs.connect())

    def test_connect_bad_ca_cert(self):
        """Test calling connect with bad ca_cert."""
        self.mqtt_cs.ca_cert = "some_certificate"

        self.assertFalse(self.mqtt_cs.connect())

    def test_connect_good_ca_cert(self):
        """Test calling connect with good ca_cert."""
        self.mqtt_cs.logger.warning = MagicMock()
        self.mqtt_cs.ca_cert = self.ca_crt_path
        self.mqtt_cs._on_mqtt_disconnect = MagicMock()

        self.assertFalse(self.mqtt_cs.connect())

    def test_connect_bad_host(self):
        """Test calling connect with bad host."""
        self.mqtt_cs.host = "bad_host"
        self.mqtt_cs.ca_cert = self.ca_crt_path
        self.mqtt_cs._on_mqtt_disconnect = MagicMock()

        self.assertFalse(self.mqtt_cs.connect())

    def test_connect_timeout(self):
        """Test connect with timeout."""
        self.mqtt_cs._on_mqtt_disconnect = MagicMock()
        self.mqtt_cs.client.connect = MagicMock()
        self.mqtt_cs.logger.warning = MagicMock()
        self.mqtt_cs.timeout_interval = -1
        self.mqtt_cs.client.subscribe = MagicMock()

        self.mqtt_cs.connect()

        self.mqtt_cs.logger.warning.assert_called_once()

    def test_connect_rc_0_with_topics(self):
        """Test connect with timeout."""
        self.mqtt_cs.topics = [1, 2, 3]
        self.mqtt_cs.ca_cert = self.ca_crt_path
        self.mqtt_cs._on_mqtt_disconnect = MagicMock()
        self.mqtt_cs.client.connect = MagicMock()
        self.mqtt_cs.logger.warning = MagicMock()
        self.mqtt_cs.connected_rc = 0
        self.mqtt_cs.client.subscribe = MagicMock()

        self.mqtt_cs.connect()

        #self.assertEqual(3, self.mqtt_cs.client.subscribe.call_count)

    def test_connect_rc_1(self):
        """Test connect with return code 1."""
        self.mqtt_cs.ca_cert = self.ca_crt_path
        self.mqtt_cs._on_mqtt_disconnect = MagicMock()
        self.mqtt_cs.client.connect = MagicMock()
        self.mqtt_cs.logger.warning = MagicMock()
        self.mqtt_cs.connected_rc = 1

        self.mqtt_cs.connect()

        #self.mqtt_cs.logger.warning.assert_called_once()

    def test_connect_rc_2(self):
        """Test connect with return code 2."""
        self.mqtt_cs.ca_cert = self.ca_crt_path
        self.mqtt_cs._on_mqtt_disconnect = MagicMock()
        self.mqtt_cs.client.connect = MagicMock()
        self.mqtt_cs.logger.warning = MagicMock()
        self.mqtt_cs.connected_rc = 2
        self.mqtt_cs.connect()

        #self.mqtt_cs.logger.warning.assert_called_once()

    def test_connect_rc_3(self):
        """Test connect with return code 3."""
        self.mqtt_cs.ca_cert = self.ca_crt_path
        self.mqtt_cs._on_mqtt_disconnect = MagicMock()
        self.mqtt_cs.client.connect = MagicMock()
        self.mqtt_cs.logger.warning = MagicMock()
        self.mqtt_cs.connected_rc = 3

        self.mqtt_cs.connect()

        #self.mqtt_cs.logger.warning.assert_called_once()

    def test_connect_rc_4(self):
        """Test connect with return code 4."""
        self.mqtt_cs.ca_cert = self.ca_crt_path
        self.mqtt_cs._on_mqtt_disconnect = MagicMock()
        self.mqtt_cs.client.connect = MagicMock()
        self.mqtt_cs.logger.warning = MagicMock()
        self.mqtt_cs.connected_rc = 4

        self.mqtt_cs.connect()

        #self.mqtt_cs.logger.warning.assert_called_once()

    def test_connect_rc_5(self):
        """Test connect with return code 5."""
        self.mqtt_cs.ca_cert = self.ca_crt_path
        self.mqtt_cs._on_mqtt_disconnect = MagicMock()
        self.mqtt_cs.client.connect = MagicMock()
        self.mqtt_cs.logger.warning = MagicMock()
        self.mqtt_cs.connected_rc = 5

        self.mqtt_cs.connect()

        #self.mqtt_cs.logger.warning.assert_called_once()

    def test_connect_rc_9_invalid(self):
        """Test connect with invalid return code 9."""
        self.mqtt_cs.ca_cert = self.ca_crt_path
        self.mqtt_cs._on_mqtt_disconnect = MagicMock()
        self.mqtt_cs.client.connect = MagicMock()
        self.mqtt_cs.logger.warning = MagicMock()
        self.mqtt_cs.connected_rc = 9

        self.mqtt_cs.connect()

        #self.mqtt_cs.logger.warning.assert_called_once()

    def test_disconnect(self):
        """Test disconnect."""
        self.mqtt_cs.logger.debug = MagicMock()
        self.mqtt_cs.client = MagicMock()
        self.mqtt_cs.client.disconnect = MagicMock()

        self.mqtt_cs.disconnect()

        self.mqtt_cs.client.disconnect.assert_called_once()

    def test_publish_no_message(self):
        """Test publish with no message."""
        message = None
        self.assertFalse(self.mqtt_cs.publish(message))

    def test_publish_not_connected(self):
        """Test publish when not connected."""
        message = Message("some_topic")

        self.assertFalse(self.mqtt_cs.publish(message))

    def test_publish_return_code_success(self):
        """Test publish with return code success."""
        self.mqtt_cs.connected = True
        message_info = mqtt.MQTTMessageInfo(1)
        self.mqtt_cs.client.publish = MagicMock(return_value=message_info)
        message = Message("some_topic")

        self.assertTrue(self.mqtt_cs.publish(message))

    def test_publish_is_published(self):
        """Test publish with is published method."""
        self.mqtt_cs.connected = True
        message_info = mqtt.MQTTMessageInfo(1)
        message_info.rc = 10
        message_info._published = True
        self.mqtt_cs.client.publish = MagicMock(return_value=message_info)
        message = Message("some_topic")

        #self.assertTrue(self.mqtt_cs.publish(message))

    def test_publish_fail_to_publish(self):
        """Test publish failing to publish message."""
        self.mqtt_cs.connected = True
        message_info = mqtt.MQTTMessageInfo(1)
        message_info.rc = 10
        self.mqtt_cs.client.publish = MagicMock(return_value=message_info)
        message = Message("some_topic")

        #self.assertFalse(self.mqtt_cs.publish(message))
